/*
  QtCurve (C) Craig Drummond, 2003   Craig.Drummond@lycos.co.uk

  Based upon FreeCurve/B???Curve
  ----

 * B???curve theme engine
 * Copyright (C) 2001 R?d H?t, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Written by Owen Taylor <otaylor@r?dh?t.com>
 * modified by Alexander Larsson <alexl@r?dh?t.com>
 */

#include <gmodule.h>
#include <gtk/gtk.h>
#include <gtk/gtkstyle.h>
#include <gdk/gdktypes.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qtcurve.h"
#include "qt_settings.h"

#ifdef GTK2_VER
#define QT_CURVE_STYLE_VAR QtCurveStyle *qtcurve_style = QTCURVE_STYLE (style);
#define QT_STYLE style
#else
#define QT_CURVE_STYLE_VAR QtCurveStyle *qtcurve_style = style->engine_data;
#define QT_STYLE style->klass
#endif

#define RADIO_SIZE 12
#define CHECK_SIZE 13
#define SCALE_SIZE 5

#if 0
static void display_widget(GtkWidget *widget, int level)
{
    if(level>=0)
    {
        printf("%s(%s) ", widget ? gtk_type_name(GTK_WIDGET_TYPE(widget)) : "NULL", widget && widget->name ? widget->name : "NULL");
        if(widget && widget->parent)
            display_widget(widget->parent, --level);
        else
            printf("\n");
    }
    else
        printf("\n");
}
#endif

static GtkStyleClass *parent_class;
static QtCurveGlobalData global;

#ifndef GTK2_VER
typedef struct _GtkBorder      GtkBorder;
struct _GtkBorder
{
  gint left;
  gint right;
  gint top;
  gint bottom;
};
#endif

static GtkRequisition default_option_indicator_size = { 7, 13 };
static GtkBorder default_option_indicator_spacing = { 7, 5, 2, 2 };

#define DETAIL(xx)   ((detail) && (!strcmp(xx, detail)))

static void
shade (GdkColor * a, GdkColor * b, float k);

static gboolean use_button_color(const gchar *detail)
{
    return detail && ( 0==strcmp(detail, "optionmenu") ||
                       0==strcmp(detail, "button") ||
                       /*0==strcmp(detail, "buttondefault") ||*/
                       0==strcmp(detail, "togglebutton") ||
                       0==strcmp(detail, "hscale") ||
                       0==strcmp(detail, "vscale") ||
                       0==strcmp(detail, "spinbutton") ||
                       0==strcmp(detail, "spinbutton_up") ||
                       0==strcmp(detail, "spinbutton_down") ||
                       0==strcmp(detail, "slider") ||
                       0==strcmp(detail, "vscrollbar") ||
                       0==strcmp(detail, "hscrollbar") );
}

#define QT_COL_EQ(A, B) (abs(A-B)<(3<<8))

#define QT_CUSTOM_COLOR_BUTTON(det, style) \
    (det && (0==strcmp(det, "button")||0==strcmp(det, "togglebutton")/*||0==strcmp(det, "buttondefault")*/) && style && \
    !(QT_COL_EQ(qt_settings.colors[COLOR_WINDOW].red, (style->bg[GTK_STATE_NORMAL].red)) && \
      QT_COL_EQ(qt_settings.colors[COLOR_WINDOW].green, (style->bg[GTK_STATE_NORMAL].green)) && \
      QT_COL_EQ(qt_settings.colors[COLOR_WINDOW].blue, (style->bg[GTK_STATE_NORMAL].blue))))

static void shade_colors(GdkColor *base, GdkColor *vals)
{
  double shades[] = {1.15, 1.05, 0.896, 0.79, 0.768, 0.4 };
  int i;

  for(i=0; i<6; ++i)
    shade (base, &vals[i], (shades[i]-0.7) * 1.0 + 0.7);

  vals[6]=*base;
}

static GdkGC *
realize_color (GtkStyle * style,
               GdkColor * color);

#define GEN_GCS(style, colors, gcs) \
{ \
  int i; \
 \
  for(i=0;i<7;++i) \
    gcs[i] = realize_color (style, &colors[i]); \
}

#define RELEASE_GCS(gcs) \
{ \
  int i; \
 \
  for(i=0; i<7;++i) \
    gtk_gc_release (gcs[i]); \
}

static void draw_lines(GdkWindow *window, int rx, int ry, int rwidth, int rheight, gboolean horiz, int n_lines, int offset, GdkGC *gcs[], GdkRectangle *area)
{
    int space = (n_lines*2)+(n_lines-1),
        x = horiz ? rx : rx+((rwidth-space)/2),
        y = horiz ? ry+((rheight-space)/2) : ry,
        x2 = rx + rwidth-1,
        y2 = ry + rheight-1,
        i; 

    if (area)
    {
      gdk_gc_set_clip_rectangle (gcs[0], area);
      gdk_gc_set_clip_rectangle (gcs[4], area);
    }
    if (horiz)
    {
        for(i=0; i<space; i+=3)
            gdk_draw_line(window, gcs[0], x+offset, y+i, x2-offset, y+i);
            
        for(i=1; i<space; i+=3)
             gdk_draw_line(window, gcs[4], x+offset, y+i, x2-offset, y+i);
    }   
    else
    {
        for(i=0; i<space; i+=3)
            gdk_draw_line(window, gcs[0], x+i, y+offset, x+i, y2-offset);

        for(i=1; i<space; i+=3)
            gdk_draw_line(window, gcs[4], x+i, y+offset, x+i, y2-offset);
    }
    if (area)
    {
      gdk_gc_set_clip_rectangle (gcs[0], NULL);
      gdk_gc_set_clip_rectangle (gcs[4], NULL);
    }
}

static void
draw_check (GtkStyle      *style,
            GdkWindow     *window,
            GtkStateType   state_type,
            GtkShadowType  shadow_type,
            GdkRectangle  *area,
            GtkWidget     *widget,
#ifdef GTK2_VER
            const
#endif
                gchar   *detail,
            gint           x,
            gint           y,
            gint           width,
            gint           height)
{
  QT_CURVE_STYLE_VAR
  gboolean mnu=DETAIL("check");

  x += (width - CHECK_SIZE)/2;
  y += (height - CHECK_SIZE)/2;

  if(!mnu)
  {
    GdkPoint shadow[3]={ {x+1, y+CHECK_SIZE-1}, {x+1, y+1}, {x+CHECK_SIZE-2, y+1} };

    if (area)
    {
      gdk_gc_set_clip_rectangle (style->base_gc[state_type], area);
      gdk_gc_set_clip_rectangle (qtcurve_style->button_gc[5], area);
      gdk_gc_set_clip_rectangle (qtcurve_style->button_gc[4], area);
    }

    gdk_draw_rectangle(window, style->base_gc[state_type], TRUE, x, y, CHECK_SIZE-1, CHECK_SIZE-1);
    gdk_draw_lines(window, qtcurve_style->button_gc[4], shadow, 3);
    gdk_draw_rectangle(window, qtcurve_style->button_gc[5], FALSE, x, y, CHECK_SIZE-1, CHECK_SIZE-1);
  }

  if (shadow_type == GTK_SHADOW_IN)
  {
            int i;
      GdkPoint check[3]={ {x+3, y+4}, {x+5, y+6}, {x+9, y+2} }; /* , {x+9, y+5}, {x+5, y+9}, {x+3, y+7}}; */

      if (area)
        gdk_gc_set_clip_rectangle (style->text_gc[state_type], area);

      for(i=0; i<3; i++)
      {
          check[0].y++;
          check[1].y++;
          check[2].y++;
          gdk_draw_lines(window, style->text_gc[state_type], check, 3);
      }
      if (area)
        gdk_gc_set_clip_rectangle (style->text_gc[state_type], NULL);
  }
  if (!mnu && area)
  {
    gdk_gc_set_clip_rectangle (style->base_gc[state_type], NULL);
    gdk_gc_set_clip_rectangle (qtcurve_style->button_gc[5], NULL);
    gdk_gc_set_clip_rectangle (qtcurve_style->button_gc[4], NULL);
  }
}

static void
draw_option (GtkStyle      *style,
             GdkWindow     *window,
             GtkStateType   state_type,
             GtkShadowType  shadow_type,
             GdkRectangle  *area,
             GtkWidget     *widget,
#ifdef GTK2_VER
             const
#endif
                 gchar   *detail,
             gint           x,
             gint           y,
             gint           width,
             gint           height)
{
  QT_CURVE_STYLE_VAR
  gboolean mnu=DETAIL ("option");

  x += (width - RADIO_SIZE)/2;
  y += (height - RADIO_SIZE)/2;

  if(!mnu)
  {
    GdkPoint outer[24]={ {x, y+7}, {x, y+4}, {x+1, y+3}, {x+1, y+2}, {x+2, y+1}, {x+3, y+1},
                         {x+4, y}, {x+7, y}, {x+8, y+1}, {x+9, y+1}, {x+10, y+2}, {x+10, y+3},
                         {x+11, y+4}, {x+11, y+7}, {x+10, y+8}, {x+10, y+9}, {x+9, y+10}, {x+8, y+10},
                         {x+7, y+11}, {x+4, y+11}, {x+3, y+10}, {x+2, y+10}, {x+1, y+9}, {x+1, y+8} },
             shadow[11]={ {x+2, y+9}, {x+2, y+8}, {x+1, y+7}, {x+1, y+4}, {x+2, y+3},
                         {x+2, y+2}, {x+3, y+2}, {x+4, y+1}, {x+7, y+1}, {x+8, y+2}, {x+9, y+2}};

    if (area)
    {
      gdk_gc_set_clip_rectangle (style->base_gc[state_type], area);
      gdk_gc_set_clip_rectangle (qtcurve_style->button_gc[5], area);
      gdk_gc_set_clip_rectangle (qtcurve_style->button_gc[4], area);
    }

    gdk_draw_rectangle(window, style->base_gc[state_type], TRUE, x+2, y+2, RADIO_SIZE-4, RADIO_SIZE-3);
    gdk_draw_line(window, style->base_gc[state_type], x+RADIO_SIZE-2, y+4, x+RADIO_SIZE-2, y+7);
    gdk_draw_lines(window, qtcurve_style->button_gc[5], outer, 24);
    gdk_draw_lines(window, qtcurve_style->button_gc[4], shadow, 11);
  }

  if (shadow_type == GTK_SHADOW_IN)
  {
      if (area)
          gdk_gc_set_clip_rectangle (style->text_gc[state_type], area);

      gdk_draw_rectangle(window, style->text_gc[state_type], TRUE, x+5, y+4, 2, 4);
      gdk_draw_rectangle(window, style->text_gc[state_type], TRUE, x+4, y+5, 4, 2);

      if (area)
          gdk_gc_set_clip_rectangle (style->text_gc[state_type], NULL);
  }

  if (!mnu && area)
  {
    gdk_gc_set_clip_rectangle (style->base_gc[state_type], NULL);
    gdk_gc_set_clip_rectangle (qtcurve_style->button_gc[5], NULL);
    gdk_gc_set_clip_rectangle (qtcurve_style->button_gc[4], NULL);
  }
}

static void draw_box (GtkStyle      *style,
		      GdkWindow     *window,
		      GtkStateType   state_type,
		      GtkShadowType  shadow_type,
		      GdkRectangle  *area,
		      GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
		      gchar         *detail,
		      gint           x,
		      gint           y,
		      gint           width,
		      gint           height);
static void draw_shadow (GtkStyle        *style,
			 GdkWindow       *window,
			 GtkStateType     state_type,
			 GtkShadowType    shadow_type,
			 GdkRectangle    *area,
			 GtkWidget       *widget,
#ifdef GTK2_VER
                      const
#endif
			 gchar           *detail,
			 gint             x,
			 gint             y,
			 gint             width,
			 gint             height);
#ifndef GTK2_VER
static void real_draw_slider (GtkStyle      *style,
			      GdkWindow     *window,
			      GtkStateType   state_type,
			      GtkShadowType  shadow_type,
			      GdkRectangle  *area,
			      GtkWidget     *widget,
			      gchar         *detail,
			      gint           x,
			      gint           y,
			      gint           width,
			      gint           height,
			      GtkOrientation orientation);
#endif

static int is_on_a_button(GtkWidget *w, int level)
{
    if(w)
    {
        if(GTK_IS_BUTTON(w) && (!(GTK_IS_RADIO_BUTTON(w) || GTK_IS_CHECK_BUTTON(w) || GTK_IS_OPTION_MENU(w))))
            return 1;
        else if (level<3)
            return is_on_a_button(w->parent, level++);
    }

    return 0;
}

#define IS_MENU_ITEM(WIDGET) (WIDGET && (GTK_IS_MENU_ITEM(WIDGET) || (WIDGET->parent && GTK_IS_MENU_ITEM(WIDGET->parent))))

#ifdef GTK2_VER
static void
draw_layout(GtkStyle *style,
            GdkWindow       *window,
            GtkStateType     state_type,
            gboolean         use_text,
            GdkRectangle    *area,
            GtkWidget       *widget,
            const gchar     *detail,
            gint             x,
            gint             y,
            PangoLayout     *layout)
{
    int selectedText=IS_MENU_ITEM(widget) && GTK_STATE_PRELIGHT==state_type;
    parent_class->draw_layout(style, window, selectedText ? GTK_STATE_SELECTED : state_type, use_text || selectedText, area, widget, detail, x, y, layout);
}
#else
static void         draw_string(GtkStyle * style,
                                GdkWindow * window,
                                GtkStateType state_type,
                                GdkRectangle * area,
                                GtkWidget * widget,
                                gchar * detail,
                                gint x,
                                gint y,
                                const gchar * string)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if(is_on_a_button(widget, 0))
  {
    int offset=0;
    GdkGC *gc=GTK_STATE_PRELIGHT==state_type ? style->text_gc[GTK_STATE_NORMAL] : style->text_gc[state_type];

    offset=GTK_STATE_ACTIVE==state_type ? 1 : 0;

    if (area)
    {
      gdk_gc_set_clip_rectangle(style->white_gc, area);
      gdk_gc_set_clip_rectangle(gc, area);
    }
    if (state_type == GTK_STATE_INSENSITIVE)
      gdk_draw_string(window, style->font, style->white_gc,
                      x + 1 + offset,
                      y + 1 + offset, string);
    gdk_draw_string(window, style->font, gc,
                      x + offset,
                      y + offset, string);
    if (area)
    {
      gdk_gc_set_clip_rectangle(style->white_gc, NULL);
      gdk_gc_set_clip_rectangle(gc, NULL);
    }
  }
  else
    parent_class->draw_string (style, window, IS_MENU_ITEM(widget) && GTK_STATE_PRELIGHT==state_type ? GTK_STATE_SELECTED : state_type, area, widget, detail, x, y, string);
}
#endif

static gboolean 
sanitize_size (GdkWindow      *window,
	       gint           *width,
	       gint           *height)
{
  gboolean set_bg = FALSE;

  if ((*width == -1) && (*height == -1))
    {
#ifdef GTK2_VER
      set_bg = GDK_IS_WINDOW (window);
#endif
      gdk_window_get_size (window, width, height);
    }
  else if (*width == -1)
    gdk_window_get_size (window, width, NULL);
  else if (*height == -1)
    gdk_window_get_size (window, NULL, height);

  return set_bg;
}

static void
option_menu_get_props (GtkWidget      * widget, 
		       GtkRequisition *indicator_size,
		       GtkBorder      *indicator_spacing)
{
  GtkRequisition *tmp_size = NULL;
  GtkBorder *tmp_spacing = NULL;

#ifdef GTK2_VER
  if (widget)
    gtk_widget_style_get (widget, 
			  "indicator_size", &tmp_size,
			  "indicator_spacing", &tmp_spacing,
			  NULL);
#endif
  if (tmp_size)
    {
      *indicator_size = *tmp_size;
      g_free (tmp_size);
    }
  else
    *indicator_size = default_option_indicator_size;

  if (tmp_spacing)
    {
      *indicator_spacing = *tmp_spacing;
      g_free (tmp_spacing);
    }
  else
    *indicator_spacing = default_option_indicator_spacing;
}

static void
rgb_to_hls (gdouble *r,
            gdouble *g,
            gdouble *b)
{
  gdouble min;
  gdouble max;
  gdouble red;
  gdouble green;
  gdouble blue;
  gdouble h, l, s;
  gdouble delta;
  
  red = *r;
  green = *g;
  blue = *b;
  
  if (red > green)
    {
      if (red > blue)
        max = red;
      else
        max = blue;
      
      if (green < blue)
        min = green;
      else
        min = blue;
    }
  else
    {
      if (green > blue)
        max = green;
      else
        max = blue;
      
      if (red < blue)
        min = red;
      else
        min = blue;
    }
  
  l = (max + min) / 2;
  s = 0;
  h = 0;
  
  if (max != min)
    {
      if (l <= 0.5)
        s = (max - min) / (max + min);
      else
        s = (max - min) / (2 - max - min);
      
      delta = max -min;
      if (red == max)
        h = (green - blue) / delta;
      else if (green == max)
        h = 2 + (blue - red) / delta;
      else if (blue == max)
        h = 4 + (red - green) / delta;
      
      h *= 60;
      if (h < 0.0)
        h += 360;
    }
  
  *r = h;
  *g = l;
  *b = s;
}

static void
hls_to_rgb (gdouble *h,
            gdouble *l,
            gdouble *s)
{
  gdouble hue;
  gdouble lightness;
  gdouble saturation;
  gdouble m1, m2;
  gdouble r, g, b;
  
  lightness = *l;
  saturation = *s;
  
  if (lightness <= 0.5)
    m2 = lightness * (1 + saturation);
  else
    m2 = lightness + saturation - lightness * saturation;
  m1 = 2 * lightness - m2;
  
  if (saturation == 0)
    {
      *h = lightness;
      *l = lightness;
      *s = lightness;
    }
  else
    {
      hue = *h + 120;
      while (hue > 360)
        hue -= 360;
      while (hue < 0)
        hue += 360;
      
      if (hue < 60)
        r = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
        r = m2;
      else if (hue < 240)
        r = m1 + (m2 - m1) * (240 - hue) / 60;
      else
        r = m1;
      
      hue = *h;
      while (hue > 360)
        hue -= 360;
      while (hue < 0)
        hue += 360;
      
      if (hue < 60)
        g = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
        g = m2;
      else if (hue < 240)
        g = m1 + (m2 - m1) * (240 - hue) / 60;
      else
        g = m1;
      
      hue = *h - 120;
      while (hue > 360)
        hue -= 360;
      while (hue < 0)
        hue += 360;
      
      if (hue < 60)
        b = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
        b = m2;
      else if (hue < 240)
        b = m1 + (m2 - m1) * (240 - hue) / 60;
      else
        b = m1;
      
      *h = r;
      *l = g;
      *s = b;
    }
}

static void
shade (GdkColor * a, GdkColor * b, float k)
{
  gdouble red;
  gdouble green;
  gdouble blue;
  
  red = (gdouble) a->red / 65535.0;
  green = (gdouble) a->green / 65535.0;
  blue = (gdouble) a->blue / 65535.0;
  
  rgb_to_hls (&red, &green, &blue);
  
  green *= k;
  if (green > 1.0)
    green = 1.0;
  else if (green < 0.0)
    green = 0.0;
  
  blue *= k;
  if (blue > 1.0)
    blue = 1.0;
  else if (blue < 0.0)
    blue = 0.0;
  
  hls_to_rgb (&red, &green, &blue);
  
  b->red = red * 65535.0;
  b->green = green * 65535.0;
  b->blue = blue * 65535.0;
}

#ifndef GTK2_VER
void gdk_rgb_find_color (GdkColormap *colormap, GdkColor *color);

void
gdk_rgb_find_color (GdkColormap *colormap, GdkColor *color)
{
  color->pixel = gdk_rgb_xpixel_from_rgb (((color->red >> 8) << 16) |
					  ((color->green >> 8) << 8) |
					  (color->blue >> 8));
}
#endif

static void
draw_hgradient (GdkDrawable *drawable, GdkGC *gc, GdkColormap *colormap,
		int x, int y, int width, int height,
		GdkColor *top_color, GdkColor *bottom_color)
{
  int i;
  GdkColor col;
  int dr, dg, db;
  GdkGCValues old_values;

  col = *top_color;
  dr = (bottom_color->red - top_color->red) / height;
  dg = (bottom_color->green - top_color->green) / height;
  db = (bottom_color->blue - top_color->blue) / height;
  
  gdk_gc_get_values (gc, &old_values);
  
  for (i = 0; i < height; i++) 
    /*for(i=height-1; i>=0; i--)*/
    {
      gdk_rgb_find_color (colormap,
			  &col);
      
      gdk_gc_set_foreground (gc, &col);
      gdk_draw_line (drawable, gc,
		     x, y + i,
		     x + width - 1, y + i);
					   
      col.red += dr;
      col.green += dg;
      col.blue += db;
    }

  gdk_gc_set_foreground (gc, &old_values.foreground);
}

static void
draw_vgradient (GdkDrawable *drawable, GdkGC *gc, GdkColormap *colormap,
		int x, int y, int width, int height,
		GdkColor *left_color, GdkColor *right_color)
{
  int i;
  GdkColor col;
  int dr, dg, db;
  GdkGCValues old_values;

  col = *left_color;
  dr = (right_color->red - left_color->red) / width;
  dg = (right_color->green - left_color->green) / width;
  db = (right_color->blue - left_color->blue) / width;
  
  gdk_gc_get_values (gc, &old_values);
  
  for (i = 0; i < width; i++)
    /*for(i=width-1; i>=0; i--)*/
    {
      gdk_rgb_find_color (colormap,
			  &col);
      
      gdk_gc_set_foreground (gc, &col);
      gdk_draw_line (drawable, gc,
		     x + i, y,
		     x + i, y + height - 1);
					   
      col.red += dr;
      col.green += dg;
      col.blue += db;
    }

  gdk_gc_set_foreground (gc, &old_values.foreground);
}

static void
arrow_draw_hline (GdkWindow     *window,
		  GdkGC         *gc,
		  int            x1,
		  int            x2,
		  int            y,
		  gboolean       last)
{
  if (x2 - x1 < 7 && !last) /* 7 to get garretts pixels, otherwise 6 */
    gdk_draw_line (window, gc, x1, y, x2, y);
  else if (last)
    {
      if (x2 - x1 <= 7)
	{
	  gdk_draw_line (window, gc, x1+1, y, x1+1, y);
	  gdk_draw_line (window, gc, x2-1, y, x2-1, y); 
	}
      else
	{
	  gdk_draw_line (window, gc, x1+2, y, x1+2, y);
	  gdk_draw_line (window, gc, x2-2, y, x2-2, y);
	}
    }
  else
    {
      gdk_draw_line (window, gc, x1, y, x1+2, y);
      gdk_draw_line (window, gc, x2-2, y, x2, y); 
    }
}

static void
arrow_draw_vline (GdkWindow     *window,
		  GdkGC         *gc,
		  int            y1,
		  int            y2,
		  int            x,
		  gboolean       last)
{
  if (y2 - y1 < 7 && !last) /* 7 to get garretts pixels */
    gdk_draw_line (window, gc, x, y1, x, y2);
  else if (last)
    {
      gdk_draw_line (window, gc, x, y1+2, x, y1+2);
      gdk_draw_line (window, gc, x, y2-2, x, y2-2); 
    }
  else
    {
      gdk_draw_line (window, gc, x, y1, x, y1+2);
      gdk_draw_line (window, gc, x, y2-2, x, y2); 
    }
}

static void
draw_arrow (GdkWindow     *window,
	    GdkGC         *gc,
	    GdkRectangle  *area,
	    GtkArrowType   arrow_type,
	    gint           x,
	    gint           y,
	    gint           width,
	    gint           height)
{
  gint i, j;

  if (area)
    gdk_gc_set_clip_rectangle (gc, area);

  if (arrow_type == GTK_ARROW_DOWN)
    {
      for (i = 0, j = -1; i < height; i++, j++)
	arrow_draw_hline (window, gc, x + j, x + width - j - 1, y + i, i == 0);
	  
    }
  else if (arrow_type == GTK_ARROW_UP)
    {
      for (i = height - 1, j = -1; i >= 0; i--, j++)
	arrow_draw_hline (window, gc, x + j, x + width - j - 1, y + i, i == height - 1);
    }
  else if (arrow_type == GTK_ARROW_LEFT)
    {
      for (i = width - 1, j = -1; i >= 0; i--, j++)
	arrow_draw_vline (window, gc, y + j, y + height - j - 1, x + i, i == width - 1);
    }
  else if (arrow_type == GTK_ARROW_RIGHT)
    {
      for (i = 0, j = -1; i < width; i++, j++)
	arrow_draw_vline (window, gc, y + j, y + height - j - 1,  x + i, i == 0);
    }

  if (area)
    gdk_gc_set_clip_rectangle (gc, NULL);
}

static void
calculate_arrow_geometry (GtkArrowType  arrow_type,
			  gint         *x,
			  gint         *y,
			  gint         *width,
			  gint         *height)
{
  gint w = *width;
  gint h = *height;
  
  switch (arrow_type)
    {
    case GTK_ARROW_UP:
    case GTK_ARROW_DOWN:
      w += (w % 2) - 1;
      h = (w / 2 + 1) + 1;

      if (h > *height)
	{
	  h = *height;
	  w = 2 * (h - 1) - 1;
	}
      
      if (arrow_type == GTK_ARROW_DOWN)
	{
	  if (*height % 2 == 1 || h % 2 == 0)
	    *height += 1;
	}
      else
	{
	  if (*height % 2 == 0 || h % 2 == 0)
	    *height -= 1;
	}
      break;

    case GTK_ARROW_RIGHT:
    case GTK_ARROW_LEFT:
      h += (h % 2) - 1;
      w = (h / 2 + 1) + 1; 
      
      if (w > *width)
	{
	  w = *width;
	  h = 2 * (w - 1) - 1;
	}
      
      if (arrow_type == GTK_ARROW_RIGHT)
	{
	  if (*width % 2 == 1 || w % 2 == 0)
	    *width += 1;
	}
      else
	{
	  if (*width % 2 == 0 || w % 2 == 0)
	    *width -= 1;
	}
      break;
      
    default:
      /* should not be reached */
      break;
    }

  *x += (*width - w) / 2;
  *y += (*height - h) / 2;
  *height = h;
  *width = w;
}

static void
draw_tab (GtkStyle      *style,
          GdkWindow     *window,
          GtkStateType   state_type,
          GtkShadowType  shadow_type,
          GdkRectangle  *area,
          GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
          gchar         *detail,
          gint           x,
          gint           y,
          gint           width,
          gint           height);

static void
qtcurve_draw_arrow (GtkStyle      *style,
		       GdkWindow     *window,
		       GtkStateType   state,
		       GtkShadowType  shadow,
		       GdkRectangle  *area,
		       GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
		       gchar         *detail,
		       GtkArrowType   arrow_type,
		       gboolean       fill,
		       gint           x,
		       gint           y,
		       gint           width,
		       gint           height)
{
#ifdef GTK2_VER
  QtCurveStyle *qtcurve_style = QTCURVE_STYLE (style);
#else
  QtCurveStyle *qtcurve_style = style->engine_data;
#endif

  if(DETAIL("arrow"))
      draw_tab(style, window, state, shadow, area, widget, detail, x, y, width, height);
  else
  {
  gint original_width, original_x;
  int is_spinbutton = DETAIL ("spinbutton");

  sanitize_size (window, &width, &height);

#ifndef GTK2_VER 
  if (DETAIL("hscrollbar") || DETAIL("vscrollbar") || is_spinbutton)
    {
      const char *box_detail = detail;
      int box_x = x, box_y = y, box_width = width, box_height = height;
      
      if (is_spinbutton)
	{
	  box_detail = (arrow_type == GTK_ARROW_DOWN)?"spinbutton_down":"spinbutton_up";
	  box_x -= 2;
	  box_width += 2*2;
	  if (arrow_type == GTK_ARROW_UP)
	    {
	      box_y -= 2;
	      box_height += 2;
	    }
	  else
	    {
	      box_y -= 0;
	      box_height += 2;
	    }
	}

      if (is_spinbutton && shadow == GTK_SHADOW_ETCHED_IN)
	{
	  shadow = GTK_SHADOW_OUT;
	  state = GTK_STATE_INSENSITIVE;
	}

      draw_box (style, window, state, shadow, area, widget, box_detail, box_x, box_y, box_width, box_height);

      if (is_spinbutton)
	{
	  x += 3;
	  y += 2;

	  width -= 2*3;
	  height -= 2*2;
	}
      else
	{
	  /* scrollbar */
	  x += 4;
	  y += 4;

	  width -= 2*4;
	  height -= 2*4;
	}
    }
#else
    if (is_spinbutton && arrow_type == GTK_ARROW_UP)
        y-=1;
#endif
 
  original_width = width;
  original_x = x;

  /* Make spinbutton arrows slightly larger to get
   * the right pixels drawn */
  if (is_spinbutton)
    height += 1;
  else if (is_spinbutton || DETAIL ("menuitem"))
    {
      width = 6;
      height = 7;
    }

  calculate_arrow_geometry (arrow_type, &x, &y, &width, &height);
  
  if (detail && strcmp (detail, "menuitem") == 0)
    x = original_x + original_width - width;


  draw_arrow (window, style->text_gc[IS_MENU_ITEM(widget) && GTK_STATE_PRELIGHT==state ? GTK_STATE_SELECTED : state], area, arrow_type,
	      x, y, width, height);
  }
}

static void
draw_tab (GtkStyle      *style,
	  GdkWindow     *window,
	  GtkStateType   state_type,
	  GtkShadowType  shadow_type,
	  GdkRectangle  *area,
	  GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
	  gchar         *detail,
	  gint           x,
	  gint           y,
	  gint           width,
	  gint           height)
{
#define ARR_WIDTH  7
#define ARR_HEIGHT 5
  draw_arrow(window, style->text_gc[state_type], area, GTK_ARROW_DOWN, x+((width-ARR_WIDTH)/2), y+((height-ARR_HEIGHT)/2), ARR_WIDTH, ARR_HEIGHT);
}

static void
draw_shadow (GtkStyle        *style,
	     GdkWindow       *window,
	     GtkStateType     state_type,
	     GtkShadowType    shadow_type,
	     GdkRectangle    *area,
	     GtkWidget       *widget,
#ifdef GTK2_VER
                      const
#endif
	     gchar           *detail,
	     gint             x,
	     gint             y,
	     gint             width,
	     gint             height)
{
#ifdef GTK2_VER
  QtCurveStyle *qtcurve_style = QTCURVE_STYLE (style);
#else
  QtCurveStyle *qtcurve_style = style->engine_data;
#endif
  GdkGC *outer_gc = NULL,
        *gc1 = NULL,
        *gc2 = NULL;

  gboolean use_bc = use_button_color(detail),
           custom_c = FALSE,
           frame=!detail || 0==strcmp(detail, "frame") || 0==strcmp(detail, "entry"),
           menu=DETAIL("menu");
  int      dark=use_bc && global.border_button ? 4 : 5;

  GdkGC    *btn_gcs[7];

  if(use_bc && QT_CUSTOM_COLOR_BUTTON(detail, style))
  {
    GdkColor btn_colors[7];
    custom_c=TRUE;
    shade_colors(&(style->bg[state_type]), btn_colors);
    GEN_GCS(style, btn_colors, btn_gcs);
  }
  else
    memcpy(btn_gcs, qtcurve_style->button_gc, 7 * sizeof(GdkGC *));

  sanitize_size (window, &width, &height);
  
#ifdef GTK2_VER /* CPD - THIS WAS IN GTK2 VER, and NOT GTK1 ??? */
  if (DETAIL("slider") && widget && GTK_IS_RANGE (widget))
    {
      GtkAdjustment *adj = GTK_RANGE (widget)->adjustment;

      if (adj->value <= adj->lower &&
	  (GTK_RANGE (widget)->has_stepper_a ||
	   GTK_RANGE (widget)->has_stepper_b))
	{
	  if (GTK_IS_VSCROLLBAR (widget))
	    {
	      height += 1;
	      y -= 1;
	    }
	  else if (GTK_IS_HSCROLLBAR (widget))
	    {
	      width += 1;
	      x -= 1;
	    }
	}
      if (adj->value >= adj->upper - adj->page_size &&
	  (GTK_RANGE (widget)->has_stepper_c ||
	   GTK_RANGE (widget)->has_stepper_d))
	{
	  if (GTK_IS_VSCROLLBAR (widget))
	    {
	      height += 1;
	    }
	  else if (GTK_IS_HSCROLLBAR (widget))
	    {
	      width += 1;
	    }
	}
    }
#endif

    if (menu || (detail &&
      ((strcmp (detail, "button") == 0) ||
       (strcmp (detail, "togglebutton") == 0) ||
       (strcmp (detail, "optionmenu") == 0) ||
       (strcmp (detail, "slider") == 0) ||
       (strcmp (detail, "vscale") == 0) ||
       (strcmp (detail, "hscale") == 0) ||
       (strcmp (detail, "vscrollbar") == 0) ||
       (strcmp (detail, "hscrollbar") == 0))))
    outer_gc = qtcurve_style->gray_gc[5];
  else if (DETAIL ("menuitem"))
    outer_gc = qtcurve_style->spot3_gc;
  else
    outer_gc = use_bc ? btn_gcs[5] : qtcurve_style->gray_gc[5];
    
  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      return;
    case GTK_SHADOW_IN:
    case GTK_SHADOW_ETCHED_IN:
      if (frame)
      {
          gc1 = qtcurve_style->gray_gc[0];
          gc2 = qtcurve_style->gray_gc[dark];
      }
      else
      {
          gc1 = use_bc ? btn_gcs[0] : qtcurve_style->gray_gc[0];
          gc2 = use_bc ? btn_gcs[dark] : qtcurve_style->gray_gc[dark];
      }
      break;
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_OUT:
      if (DETAIL ("menuitem"))
      {
	  gc1 = qtcurve_style->spot2_gc;
	  gc2 = qtcurve_style->spot1_gc;
      }
      else if (frame)
      {
          gc1 = qtcurve_style->gray_gc[5];
          gc2 = qtcurve_style->gray_gc[0];
      }
      else
      {
          gc1 = use_bc ? btn_gcs[dark] : qtcurve_style->gray_gc[4];
          gc2 = use_bc ? btn_gcs[0] : qtcurve_style->gray_gc[0];
      }
      break;
    }
  
  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, area);
      gdk_gc_set_clip_rectangle (gc2, area);
      if (shadow_type == GTK_SHADOW_IN || 
          shadow_type == GTK_SHADOW_OUT)
	{
	  gdk_gc_set_clip_rectangle (outer_gc, area);
	  gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[5], area);
	}
    }
  
  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      break;
    case GTK_SHADOW_IN:
    case GTK_SHADOW_OUT:
      if (widget && GTK_WIDGET_HAS_DEFAULT (widget) && DETAIL("button"))
	gdk_draw_rectangle (window, use_bc ? btn_gcs[5] : qtcurve_style->gray_gc[5], FALSE,
			    x, y, width - 1, height - 1);
      else if (frame || menu)
      {
          if (menu && global.border_frame)
          {
              gdk_draw_rectangle (window, outer_gc, FALSE, x, y, width - 1, height - 1);
              x++; y++; width-=2; height-=2;
          }

          gdk_draw_line (window, gc2, x,         y,          x+width-2, y);
          gdk_draw_line (window, gc2, x,         y,          x,         y+height-2);
          gdk_draw_line (window, gc1, x,         y+height-1, x+width-1, y+height-1);
          gdk_draw_line (window, gc1, x+width-1, y+height-1, x+width-1, y);
          break;
      }
	gdk_draw_rectangle (window, outer_gc, FALSE,
			    x, y, width - 1, height - 1);

      if(GTK_SHADOW_IN==shadow_type && global.border_button)
          break;

      if (detail &&
	  ((strcmp (detail, "button") == 0) ||
	   (strcmp (detail, "togglebutton") == 0) ||
	   (strcmp (detail, "handlebox") == 0) ||
	   (strcmp (detail, "optionmenu") == 0) ||
	   (strcmp (detail, "notebook") == 0) ||
	   (strcmp (detail, "slider") == 0) ||
	   (strcmp (detail, "spinbutton_up") == 0) ||
	   (strcmp (detail, "spinbutton_down") == 0) ||
	   (strcmp (detail, "menuitem") == 0) ||
           (strcmp (detail, "vscale") == 0) ||
           (strcmp (detail, "hscale") == 0) ||
	   (strcmp (detail, "vscrollbar") == 0) ||
	   (strcmp (detail, "hscrollbar") == 0) ||
	   (strcmp (detail, "stepper") == 0)))
	{
          if(!global.border_button)
          {
              x--; y--; width+=2; height+=2;
          }

	  
	  /* Top and left */
	  
	  if (QT_STYLE->ythickness > 0)
	    gdk_draw_line (window, gc2,
			   x + 1, y + 1, x + width - 2, y + 1);
	  if (QT_STYLE->xthickness > 0)
	    gdk_draw_line (window, gc2,
			   x + 1, y + 1, x + 1, y + height - 2);

          /* Right and bottom edge */
          if (QT_STYLE->ythickness > 0)
            gdk_draw_line (window, gc1,
                           x + 1, y + height - 2, x + width - 2, y + height - 2);
          if (QT_STYLE->xthickness > 0)
            gdk_draw_line (window, gc1,
                           x + width - 2, y + 1, x + width - 2, y + height - 2);
	}      
      break;
      
    case GTK_SHADOW_ETCHED_IN:
          gdk_draw_rectangle(window, gc1, FALSE, x+1, y+1, width-2, height-2);
          gdk_draw_rectangle(window, gc2, FALSE, x, y, width-2, height-2);
      break;
    case GTK_SHADOW_ETCHED_OUT:
          gdk_draw_rectangle(window, gc2, FALSE, x+1, y+1, width-2, height-2);
          gdk_draw_rectangle(window, gc1, FALSE, x, y, width-2, height-2);
      break;
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, NULL);
      gdk_gc_set_clip_rectangle (gc2, NULL);
      if (shadow_type == GTK_SHADOW_IN || 
          shadow_type == GTK_SHADOW_OUT)
	{
	  gdk_gc_set_clip_rectangle (outer_gc, NULL);
	  gdk_gc_set_clip_rectangle (use_bc ? btn_gcs[5] : qtcurve_style->gray_gc[5], NULL);
	}
    }

  if(custom_c)
    RELEASE_GCS(btn_gcs);
}

static void
draw_box_gap (GtkStyle       *style,
              GdkWindow      *window,
              GtkStateType    state_type,
              GtkShadowType   shadow_type,
              GdkRectangle   *area,
              GtkWidget      *widget,
#ifdef GTK2_VER
                      const
#endif
              gchar          *detail,
              gint            x,
              gint            y,
              gint            width,
              gint            height,
              GtkPositionType gap_side,
              gint            gap_x,
              gint            gap_width)
{
  QT_CURVE_STYLE_VAR
  GdkGC *gc1 = NULL,
        *gc2 = NULL,
        *outer = qtcurve_style->gray_gc[5];
  int   dark=global.border_frame ? 4 : 5;

#ifdef GTK2_VER
  g_return_if_fail (GTK_IS_STYLE (style));
#endif
  g_return_if_fail (window != NULL);

  gtk_style_apply_default_background (style, window,
                                      widget && !GTK_WIDGET_NO_WINDOW (widget),
                                      state_type, area, x, y, width, height);

  sanitize_size (window, &width, &height);

  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      return;
    case GTK_SHADOW_ETCHED_IN:
    case GTK_SHADOW_IN:
      gc1 = qtcurve_style->gray_gc[dark];
      gc2 = qtcurve_style->gray_gc[0];
      break;
    case GTK_SHADOW_ETCHED_OUT:
    case GTK_SHADOW_OUT:
      gc1 = qtcurve_style->gray_gc[0];
      gc2 = qtcurve_style->gray_gc[dark];
      break;
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, area);
      gdk_gc_set_clip_rectangle (gc2, area);
      gdk_gc_set_clip_rectangle (outer, area);
    }
 
  if(global.border_frame) 
    switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
    case GTK_SHADOW_IN:
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_IN:
    case GTK_SHADOW_ETCHED_OUT:
      switch (gap_side)
        {
        case GTK_POS_TOP:
          if(gap_x > 0)
          {
              gdk_draw_line(window, gc1, x, y+1, x+gap_x+1, y+1);
              gdk_draw_line(window, gc1, x+gap_x+1, y, x+gap_x+1, y);
              gdk_draw_line(window, outer, x, y, x+gap_x, y);
          }
          else
              gdk_draw_line(window, gc1, x+1, y, x+1, y+1);
          if((width - (gap_x + gap_width)) > 0)
          {
              gdk_draw_line(window, gc1, x+gap_x+gap_width-2, y+1, x+width-2, y+1);
              gdk_draw_line(window, gc2, x+gap_x+gap_width-2, y, x+gap_x+gap_width-2, y);
              gdk_draw_line(window, outer, x+gap_x+gap_width-1, y, x+width-1, y);
          }
          else
              gdk_draw_line(window, gc2, x+width-2, y, x+width-2, y);
          gdk_draw_line(window, gc1, x+1, y+2, x+1, y + height-1);
          gdk_draw_line(window, gc2, x+1, y+height-2, x+width-2, y+height-2);
          gdk_draw_line(window, gc2, x+width-2, y+1, x+width-2, y+height-1);
          gdk_draw_line(window, outer, x, y, x, y + height-1);
          gdk_draw_line(window, outer, x, y+height-1, x+width-1, y+height-1);
          gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+height-1);
          break;
        case  GTK_POS_BOTTOM:
          if(gap_x > 0)
          {
              gdk_draw_line(window, gc1, x+gap_x+1, y+height-1, x+gap_x+1, y+height-2);
              gdk_draw_line(window, gc2, x+1, y+height-2, x+gap_x, y+height-2);
              gdk_draw_line(window, outer, x, y+height-1, x+gap_x, y+height-1);
          }
          else
              gdk_draw_line(window, gc1, x+1, y+height-1, x+1, y+height-2);
          if((width - (gap_x + gap_width)) > 0)
          {
              gdk_draw_line(window, gc2, x+gap_x+gap_width-2, y+height-2, x+width-1, y+height-2);
              gdk_draw_line(window, gc2, x+gap_x+gap_width-2, y+height-1, x+gap_x+gap_width-2, y+height-1);
              gdk_draw_line(window, outer, x+gap_x+gap_width-1, y+height-1, x+width-1, y+height-1);
          }
          else
              gdk_draw_line(window, gc2, x+width-2, y+height-1, x+width-2, y+height-1);
          gdk_draw_line(window, gc1, x+1, y+1, x+width-2, y+1);
          gdk_draw_line(window, gc1, x+1, y+1, x+1, y+height-3);
          gdk_draw_line(window, gc2, x+width-2, y+1, x+width-2, y+height-2);
          gdk_draw_line(window, outer, x, y, x+width-1, y);
          gdk_draw_line(window, outer, x, y, x, y+height-1);
          gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+height-1);
          break;
        case GTK_POS_LEFT:
          if(gap_x>0)
          {
              gdk_draw_line(window, gc1, x+1, y+1, x+1, y+gap_x+1);
              gdk_draw_line(window, gc1, x, y+gap_x+1, x, y+gap_x+1);
              gdk_draw_line(window, outer, x, y, x, y+gap_x);
          }
          else
              gdk_draw_line(window, gc1, x, y+1, x+1, y+1);
          if((height-(gap_x + gap_width)) > 0)
          {
              gdk_draw_line(window, gc1, x+1, y+gap_x+gap_width-2, x+1, y+height-1);
              gdk_draw_line(window, gc2, x, y+gap_x+gap_width-2, x, y+gap_x+gap_width-2);
              gdk_draw_line(window, outer, x, y+gap_x+gap_width-1, x, y+height-1);
          }
          else
              gdk_draw_line(window, gc2, x, y+height-2, x, y+height-2);
          gdk_draw_line(window, gc1, x+1, y+1, x+width-2, y+1);
          gdk_draw_line(window, gc2, x+width-2, y+1, x+width-2, y+height-2);
          gdk_draw_line(window, gc2, x+1, y+height-2, x+width-2, y+height-2);
          gdk_draw_line(window, outer, x, y, x+width-1, y);
          gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+height-1);
          gdk_draw_line(window, outer, x, y+height-1, x+width-1, y+height-1);
          break;
        case GTK_POS_RIGHT:
          if(gap_x>0)
          {
              gdk_draw_line(window, gc2, x+width-2, y, x+width-2, y+gap_x);
              gdk_draw_line(window, gc1, x+width-2, y+gap_x+1, x+width-1, y+gap_x+1);
              gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+gap_x);
          }
          else
              gdk_draw_line(window, gc1, x+width-2, y+1, x+width, y+1);
          if((height-(gap_x + gap_width)) > 0)
          {
              gdk_draw_line(window, gc2, x+width-2, y+gap_x+gap_width-2, x+width, y+gap_x+gap_width-2);
              gdk_draw_line(window, gc2, x+width-2, y+gap_x+gap_width-1, x+width-2, y+height-2);
              gdk_draw_line(window, outer, x+width-1, y+gap_x+gap_width-1, x+width-1, y+height-1);
          }
          else
              gdk_draw_line(window, gc2, x+width-1, y+height-2, x+width-1, y+height-2);
          gdk_draw_line(window, gc1, x+1, y+1, x+width-3, y+1);
          gdk_draw_line(window, gc1, x+1, y+1, x+1, y+height-2);
          gdk_draw_line(window, gc2, x+1, y+height-2, x+width-2, y+height-2);
          gdk_draw_line(window, outer, x, y, x+width-1, y);
          gdk_draw_line(window, outer, x, y, x, y+height-1);
          gdk_draw_line(window, outer, x, y+height-1, x+width-1, y+height-1);
          break;
        }
      }
  else
    switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
    case GTK_SHADOW_IN:
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_IN:
    case GTK_SHADOW_ETCHED_OUT:
      switch (gap_side)
        {
        case GTK_POS_TOP:
          if(gap_x > 0)
              gdk_draw_line(window, gc1, x, y, x+gap_x, y);
          if((width - (gap_x + gap_width)) > 0)
          {
              gdk_draw_line(window, gc1, x+gap_x+gap_width, y, x+width-1, y);
              gdk_draw_line(window, gc2, x+gap_x+gap_width-1, y, x+gap_x+gap_width-1, y);
          }
          gdk_draw_line(window, gc1, x, y, x, y + height-1);
          gdk_draw_line(window, gc2, x, y+height-1, x+width-1, y+height-1);
          gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+height-1);
          break;
        case  GTK_POS_BOTTOM:
          if(gap_x > 0)
              gdk_draw_line(window, gc1, x, y+height-1, x+gap_x, y+height-1);
          if((width - (gap_x + gap_width)) > 0)
          {
              gdk_draw_line(window, gc1, x+gap_x+gap_width, y+height-1, x+width-1, y+height-1);
              gdk_draw_line(window, gc2, x+gap_x+gap_width-1, y+height-1, x+gap_x+gap_width-1, y+height-1);
          }
          gdk_draw_line(window, gc1, x, y, x+width-1, y);
          gdk_draw_line(window, gc1, x, y, x, y+height-1);
          gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+height-1);
          break;
        case GTK_POS_LEFT:
          if(gap_x>0)
              gdk_draw_line(window, gc1, x, y, x, y+gap_x);
          if((height-(gap_x + gap_width)) > 0)
          {
              gdk_draw_line(window, gc1, x, y+gap_x+gap_width, x, y+height-1);
              gdk_draw_line(window, gc2, x, y+gap_x+gap_width-1, x, y+gap_x+gap_width-1);
          }
          gdk_draw_line(window, gc1, x, y, x+width-1, y);
          gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+height-1);
          gdk_draw_line(window, gc2, x, y+height-1, x+width-1, y+height-1);
          break;
        case GTK_POS_RIGHT:
          if(gap_x>0)
          {
              gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+gap_x);
              gdk_draw_line(window, gc1, x+width-1, y+gap_x, x+width-1, y+gap_x);
          }
          if((height-(gap_x + gap_width)) > 0)
              gdk_draw_line(window, gc2, x+width-1, y+gap_x+gap_width-1, x+width-1, y+height-1);
          gdk_draw_line(window, gc1, x, y, x+width-1, y);
          gdk_draw_line(window, gc1, x, y, x, y+height-1);
          gdk_draw_line(window, gc2, x, y+height-1, x+width-1, y+height-1);
          break;
        }
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, NULL);
      gdk_gc_set_clip_rectangle (gc2, NULL);
      gdk_gc_set_clip_rectangle (outer, NULL);
    }
}

static void
draw_extension (GtkStyle       *style,
                GdkWindow      *window,
                GtkStateType    state_type,
                GtkShadowType   shadow_type,
                GdkRectangle   *area,
                GtkWidget      *widget,
#ifdef GTK2_VER
                      const
#endif
                gchar          *detail,
                gint            x,
                gint            y,
                gint            width,
                gint            height,
                GtkPositionType gap_side)
{
  QT_CURVE_STYLE_VAR
  GdkGC *gc1 = NULL,
        *gc2 = NULL,
        *fill = GTK_STATE_ACTIVE==state_type ? qtcurve_style->gray_gc[2] : style->bg_gc[GTK_STATE_NORMAL],
        *outer = qtcurve_style->gray_gc[5];
  int   dark=global.border_frame ? 4 : 5;

#ifdef GTK2_VER
  g_return_if_fail (GTK_IS_STYLE (style));
#endif
  g_return_if_fail (window != NULL);

  gtk_style_apply_default_background (style, window,
                                      widget && !GTK_WIDGET_NO_WINDOW (widget),
                                      GTK_STATE_NORMAL, area, x, y, width, height);

  sanitize_size (window, &width, &height);

  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      return;
    case GTK_SHADOW_IN:
    case GTK_SHADOW_ETCHED_IN:
      gc1 = qtcurve_style->gray_gc[dark];
      gc2 = qtcurve_style->gray_gc[0];
      break;
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_OUT:
      gc1 = qtcurve_style->gray_gc[0];
      gc2 = qtcurve_style->gray_gc[dark];
      break;
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, area);
      gdk_gc_set_clip_rectangle (gc2, area);
      gdk_gc_set_clip_rectangle (fill, area);
      gdk_gc_set_clip_rectangle (outer, area);
    }

  if(global.border_frame)
    switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
    case GTK_SHADOW_IN:
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_IN:
    case GTK_SHADOW_ETCHED_OUT:
      gdk_draw_rectangle(window, fill, TRUE, x, y, width, height );
      switch (gap_side)
        {
        case GTK_POS_TOP:
          gdk_draw_line (window, gc1, x+1, y , x+1, y+height-1);
          gdk_draw_line (window, gc2, x+1, y+height-2, x+width-2, y+height-2);
          gdk_draw_line (window, gc2, x+width-2, y, x+width-2, y+height-1);
          gdk_draw_line (window, outer, x, y , x, y+height-1);
          gdk_draw_line (window, outer, x, y+height-1, x+width-1, y+height-1);
          gdk_draw_line (window, outer, x+width-1, y, x+width-1, y+height-1);
          break;
        case GTK_POS_BOTTOM:
          gdk_draw_line (window, gc1, x+1, y+1 , x+width-2, y+1);
          gdk_draw_line (window, gc1, x+1, y+1, x+1, y+height-1);
          gdk_draw_line (window, gc2, x+width-2, y+1, x+width-2, y+height-1);
          gdk_draw_line (window, outer, x, y , x+width-1, y);
          gdk_draw_line (window, outer, x, y, x, y+height-1);
          gdk_draw_line (window, outer, x+width-1, y, x+width-1, y+height-1);
          break;
        case GTK_POS_LEFT:
          gdk_draw_line (window, gc1, x, y+1, x+width-1, y+1);
          gdk_draw_line (window, gc2, x+width-2, y+1, x+width-2, y+height-2);
          gdk_draw_line (window, gc2, x, y+height-2, x+width-1, y+height-2);
          gdk_draw_line (window, outer, x, y, x+width-1, y);
          gdk_draw_line (window, outer, x+width-1, y, x+width-1, y+height-1);
          gdk_draw_line (window, outer, x, y+height-1, x+width-1, y+height-1);
          break;
        case GTK_POS_RIGHT:
          gdk_draw_line (window, gc1, x+1, y+1, x+1, y+height-2);
          gdk_draw_line (window, gc1, x, y+1, x+width-1, y+1);
          gdk_draw_line (window, gc2, x, y+height-2, x+width-1, y+height-2);
          gdk_draw_line (window, outer, x, y, x, y+height-1);
          gdk_draw_line (window, outer, x, y, x+width-1, y);
          gdk_draw_line (window, outer, x, y+height-1, x+width-1, y+height-1);
          break;
        }
    }
  else
    switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
    case GTK_SHADOW_IN:
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_IN:
    case GTK_SHADOW_ETCHED_OUT:
      gdk_draw_rectangle(window, fill, TRUE, x, y, width, height );
      switch (gap_side)
        {
        case GTK_POS_TOP:
          gdk_draw_line (window, gc1, x, y , x, y+height-1);
          gdk_draw_line (window, gc2, x, y+height-1, x+width-1, y+height-1);
          gdk_draw_line (window, gc2, x+width-1, y, x+width-1, y+height-1);
          break;
        case GTK_POS_BOTTOM:
          gdk_draw_line (window, gc1, x, y , x+width-1, y);
          gdk_draw_line (window, gc1, x, y, x, y+height-1);
          gdk_draw_line (window, gc2, x+width-1, y, x+width-1, y+height-1);
          break;
        case GTK_POS_LEFT:
          gdk_draw_line (window, gc1, x, y , x+width-1, y);
          gdk_draw_line (window, gc2, x+width-1, y, x+width-1, y+height-1);
          gdk_draw_line (window, gc2, x, y+height-1, x+width-1, y+height-1);
          break;
        case GTK_POS_RIGHT:
          gdk_draw_line (window, gc1, x, y , x, y+height-1);
          gdk_draw_line (window, gc1, x, y, x+width-1, y);
          gdk_draw_line (window, gc2, x, y+height-1, x+width-1, y+height-1);
          break;
        }
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, NULL);
      gdk_gc_set_clip_rectangle (gc2, NULL);
      gdk_gc_set_clip_rectangle (fill, NULL);
      gdk_gc_set_clip_rectangle (outer, NULL);
    }
}

static void 
draw_handle (GtkStyle      *style,
	     GdkWindow     *window,
	     GtkStateType   state_type,
	     GtkShadowType  shadow_type,
	     GdkRectangle  *area,
	     GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
	     gchar         *detail,
	     gint           x,
	     gint           y,
	     gint           width,
	     gint           height,
	     GtkOrientation orientation)
{
  QT_CURVE_STYLE_VAR
  
#ifdef GTK2_VER
  g_return_if_fail (GTK_IS_STYLE (style));
#endif  
  g_return_if_fail (window != NULL);
  
  sanitize_size (window, &width, &height);
  
  gtk_style_apply_default_background (style, window,
				      widget && !GTK_WIDGET_NO_WINDOW (widget),
				      state_type, area, x, y, width, height);

  if (detail && (!strcmp (detail, "paned") || !strcmp (detail+1, "paned")))
  {
      GdkSegment dark_lines[4]=  { { x+1,       y,          x+width-2, y          },
                                   { x+width-1, y+1,        x+width-1, y+height-2 },
                                   { x+1,       y+height-1, x+width-2, y+height-1 },
                                   { x,         y+1,        x,         y+height-2 } },
                 light_lines[2]= { { x+1,       y+1,        x+width-3, y+1        },
                                   { x+1,       y+2,        x+1,       y+height-3 } };

      if (area)
      {
        gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[0], area);
        gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[3], area);
        gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[5], area);
      }

      gdk_draw_segments(window, qtcurve_style->gray_gc[5], dark_lines, 4);
      gdk_draw_segments(window, qtcurve_style->gray_gc[0], light_lines, 2);
      if (area)
      {
        gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[0], NULL);
        gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[3], NULL);
        gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[5], NULL);
      }
  }
  else if ((DETAIL ("handlebox") && widget && GTK_IS_HANDLE_BOX (widget)) || DETAIL ("dockitem"))
  {
      draw_box(style, window, state_type, shadow_type, area, widget, "handlebox", x, y, width, height);
      draw_lines(window, x, y, width, height, height<width, 2, 4, qtcurve_style->gray_gc, area);
  }
}

static void
draw_box (GtkStyle      *style,
	  GdkWindow     *window,
	  GtkStateType   state_type,
	  GtkShadowType  shadow_type,
	  GdkRectangle  *area,
	  GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
	  gchar         *detail,
	  gint           x,
	  gint           y,
	  gint           width,
	  gint           height)
{
  QT_CURVE_STYLE_VAR
  gboolean custom_c = FALSE,
           pbar=DETAIL("bar") && GTK_IS_PROGRESS_BAR(widget);
  GdkGC    *btn_gcs[7],
           *bgnd_gc=NULL;

  if(use_button_color(detail))
  {
    if(QT_CUSTOM_COLOR_BUTTON(detail, style))
    {
      GdkColor btn_colors[7];
      custom_c=TRUE;
      shade_colors(&(style->bg[state_type]), btn_colors);
      GEN_GCS(style, btn_colors, btn_gcs);
    }
    else
      memcpy(btn_gcs, qtcurve_style->button_gc, 7 * sizeof(GdkGC *));

    bgnd_gc=state_type==GTK_STATE_PRELIGHT ? btn_gcs[1] :
            state_type==GTK_STATE_ACTIVE ? btn_gcs[3] : btn_gcs[6];
  }
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  if(detail && ( 0==strcmp(detail, "spinbutton_up") || 0==strcmp(detail, "spinbutton_down")))
  {
#ifdef GTK2_VER
      x--;
      y--;
      width+=2;
      height+=2;
#else
      height++;
      if(0==strcmp(detail, "spinbutton_down"))
      {
          y--;
          height++;
      }
#endif

      if (area)
          gdk_gc_set_clip_rectangle (bgnd_gc, area);
      gdk_draw_rectangle (window, bgnd_gc, TRUE, x+1, y+1, width-2, height-2);
      if (area)
          gdk_gc_set_clip_rectangle (bgnd_gc, NULL);

      gtk_paint_shadow (style, window, state_type, shadow_type, area, widget, detail, x+1, y+1, width-2, height-2);
  }
  else if (DETAIL("spinbutton"))
      gtk_style_apply_default_background (style, window,
                                  widget && !GTK_WIDGET_NO_WINDOW (widget),
                                  GTK_STATE_INSENSITIVE==state_type ? GTK_STATE_INSENSITIVE : GTK_STATE_NORMAL, area, x, y, width, height);

  else if(detail && ( 0==strcmp(detail, "button") ||
                 0==strcmp(detail, "togglebutton") ||
                 /*0==strcmp(detail, "spinbutton") || */
                 0==strcmp(detail, "slider") ||
                 0==strcmp(detail, "hscrollbar") ||
                 0==strcmp(detail, "vscrollbar")  ||
                 0==strcmp(detail, "hscale") ||
                 0==strcmp(detail, "vscale") ) )
  {
    if (area)
      gdk_gc_set_clip_rectangle (bgnd_gc, area);

    gdk_draw_rectangle (window, bgnd_gc, TRUE, x, y, width, height);
    if (area)
      gdk_gc_set_clip_rectangle (bgnd_gc, NULL);

    gtk_paint_shadow (style, window, state_type, shadow_type, area, widget, detail, x, y, width, height);
  }
  else if (DETAIL("buttondefault"))
  {
#ifndef GTK2_VER
    gint xmod=QT_STYLE->xthickness+2,
         ymod=QT_STYLE->ythickness+2;
#endif

    if (area)
      gdk_gc_set_clip_rectangle (style->text_gc[state_type], area);

#ifdef GTK2_VER
    gdk_draw_rectangle (window, style->text_gc[state_type], TRUE, x, y, width, height);
#else
    gdk_draw_rectangle (window, style->text_gc[state_type], TRUE, x+xmod, y+ymod, width-(2*xmod), height-(2*ymod));
#endif
    if (area)
      gdk_gc_set_clip_rectangle (style->text_gc[state_type], NULL);
  }
  else
  if (widget && DETAIL ("trough")) /* Scrollbars */
    {
      gboolean pbar=GTK_IS_PROGRESS_BAR(widget);
      GdkGC *bgnd=pbar ? qtcurve_style->gray_gc[6] : qtcurve_style->gray_gc[2],
            *gcl =pbar ? qtcurve_style->gray_gc[0] : qtcurve_style->gray_gc[4],
            *gcd =/*pbar ? qtcurve_style->gray_gc[5] : */qtcurve_style->gray_gc[5];

      if (area)
	{
	  gdk_gc_set_clip_rectangle (bgnd, area);
	  gdk_gc_set_clip_rectangle (gcl, area);
	  gdk_gc_set_clip_rectangle (gcd, area);
	}

      if (GTK_IS_HSCALE (widget) || GTK_IS_VSCALE (widget))
	{
          gtk_style_apply_default_background (style, window,
                                      widget && !GTK_WIDGET_NO_WINDOW (widget),
                                      GTK_STATE_INSENSITIVE==state_type ? GTK_STATE_INSENSITIVE : GTK_STATE_NORMAL, area, x, y, width, height);
          if GTK_IS_HSCALE (widget)
          {
              y += (height - SCALE_SIZE) / 2;
              height = SCALE_SIZE;
          }
          else
          {
              x += (width - SCALE_SIZE) / 2;
              width = SCALE_SIZE;
          }

          gdk_draw_rectangle (window, bgnd, TRUE, x, y, width, height);
          gdk_draw_rectangle (window, gcd, FALSE, x, y, width - 1, height - 1);
	  gdk_draw_line (window, gcl, x + 1, y + 1, x + width - 2, y + 1);
	  gdk_draw_line (window, gcl, x + 1, y + 1, x + 1, y + height - 2);
	}
      else
      {
          gdk_draw_rectangle (window, bgnd, TRUE, x, y, width, height);

          if(GTK_IS_PROGRESS_BAR(widget))
          {
              gdk_draw_line (window, gcd, x,         y,          x+width-2, y);
              gdk_draw_line (window, gcd, x,         y,          x,         y+height-2);
              gdk_draw_line (window, gcl, x,         y+height-1, x+width-1, y+height-1);
              gdk_draw_line (window, gcl, x+width-1, y+height-1, x+width-1, y);
          }
          else if(global.border_button)
              gdk_draw_rectangle (window, gcd, FALSE, x, y, width - 1, height - 1);
      }

      if (area)
	{
	  gdk_gc_set_clip_rectangle (bgnd, NULL);
	  gdk_gc_set_clip_rectangle (gcl, NULL);
	  gdk_gc_set_clip_rectangle (gcd, NULL);
	}
    }
  else if (widget && (detail && (0==strcmp(detail, "menubar") || 0==strcmp(detail,"toolbar") || 0==strcmp(detail,"dockitem_bin") ||
                                 0==strcmp(detail, "handlebox")))
	   && shadow_type != GTK_SHADOW_NONE)
    {
      gboolean top=FALSE, bottom=FALSE, left=FALSE, right=FALSE;

      /* Toolbars and menus */
      gtk_style_apply_default_background (style, window,
					  widget && !GTK_WIDGET_NO_WINDOW (widget),
					  state_type, area, x, y, width, height);
 
      if (area)
	{
	  gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[0], area);
	  gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[5], area);
	}


      if(0==strcmp(detail, "menubar"))
          top=bottom=TRUE;
      else if (0==strcmp(detail,"toolbar") && GTK_IS_TOOLBAR(widget))
      {
#ifdef GTK2_VER
          if(GTK_ORIENTATION_HORIZONTAL==gtk_toolbar_get_orientation (GTK_TOOLBAR (widget)))
#else
          if(GTK_ORIENTATION_HORIZONTAL== GTK_PROGRESS_BAR(widget)->orientation)
#endif
              top=bottom=TRUE;
          else   /* MUST have a handle */
              left=right=bottom=TRUE;
      }
      else if(0==strcmp(detail,"dockitem_bin"))  /* CPD: bit risky - what if only 1 item ??? */
      {
          if(width<height)
              left=right=top=TRUE;
          else
              top=bottom=TRUE;
      }
      else /* handle */
          if(width<height) /* on horiz toolbar */
              top=bottom=TRUE;
          else
              left=right=top=TRUE;

      if(top)
          gdk_draw_line (window, qtcurve_style->gray_gc[0],
		       x, y, x + width-1, y);
      if(left)
          gdk_draw_line (window, qtcurve_style->gray_gc[0],
                       x, y, x, y+height-1);
      if(bottom)
          gdk_draw_line (window, qtcurve_style->gray_gc[5],
		     x, y + height - 1, x + width-1, y + height - 1);
      if(right)
          gdk_draw_line (window, qtcurve_style->gray_gc[5],
                     x+width-1, y, x + width-1, y + height - 1);
      
      if (area)
	{
	  gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[0], NULL);
	  gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[5], NULL);
	}
    }
  else if (widget && (DETAIL("menuitem") || pbar))
    {
      GdkColor upper_color, lower_color;

      shade (&qt_settings.colors[COLOR_SELECTED], &upper_color, 0.92);
      shade (&qt_settings.colors[COLOR_SELECTED], &lower_color, 1.66);
      
      if (area)
	gdk_gc_set_clip_rectangle (style->base_gc[state_type], area);

      if(!pbar)
          draw_hgradient (window, style->base_gc[state_type], style->colormap,
	   	          x, y, width, height,
		          &upper_color, &lower_color);
      else
      {
              GtkProgressBarOrientation
#ifdef GTK2_VER
              orientation = gtk_progress_bar_get_orientation (GTK_PROGRESS_BAR (widget));
#else
              orientation = GTK_PROGRESS_BAR (widget)->orientation;
#endif

          if (orientation == GTK_PROGRESS_LEFT_TO_RIGHT ||
              orientation == GTK_PROGRESS_RIGHT_TO_LEFT)
          {
              if (width > 2)
                  draw_hgradient (window, style->base_gc[state_type], style->colormap,
                                x, y, width, height,
                                &upper_color, &lower_color);
          }
          else
          {
              if (height > 2)
                draw_vgradient (window, style->base_gc[state_type], style->colormap,
                                x, y, width, height,
                               &upper_color, &lower_color);
          }
      }

      
      if (area)
	gdk_gc_set_clip_rectangle (style->base_gc[state_type], NULL);

      
      gtk_paint_shadow (style, window, state_type, shadow_type, area, widget, "menuitem",
			x, y, width, height);
      
    }
  else if (widget && DETAIL ("optionmenu"))
    {
      GtkRequisition indicator_size;
      GtkBorder indicator_spacing;

      if (!style->bg_pixmap[state_type]
#ifdef GTK2_VER
              || GDK_IS_PIXMAP (window)
#endif
         )
        {
          if (area)
            gdk_gc_set_clip_rectangle (bgnd_gc, area);

          gdk_draw_rectangle (window, bgnd_gc, TRUE,
                              x, y, width, height);
          if (area)
            gdk_gc_set_clip_rectangle (bgnd_gc, NULL);
        }
      else
      gtk_style_apply_default_background (style, window,
					  widget && !GTK_WIDGET_NO_WINDOW (widget),
					  state_type, area, x, y, width, height);


      gtk_paint_shadow (style, window, state_type, shadow_type, area, widget, detail,
			x, y, width, height);
      
      option_menu_get_props (widget, &indicator_size, &indicator_spacing);

      if (area)
	{
	  gdk_gc_set_clip_rectangle (btn_gcs[0], area);
	  gdk_gc_set_clip_rectangle (btn_gcs[4], area);
	}

      if(!global.border_button)
      {
          y--;
          height+=2;
      }
      
      gdk_draw_line (window, btn_gcs[global.border_button ? 4 : 5],
		     x + width - (indicator_size.width + indicator_spacing.left + indicator_spacing.right) - QT_STYLE->xthickness,
#ifdef GTK2_VER
		     y + QT_STYLE->ythickness - 1,
#else
                     y + QT_STYLE->ythickness,
#endif
		     x + width - (indicator_size.width + indicator_spacing.left + indicator_spacing.right) - QT_STYLE->xthickness,
#ifdef GTK2_VER
		     y + height - QT_STYLE->ythickness);
#else
                     y + height - QT_STYLE->ythickness-2);
#endif
	
      gdk_draw_line (window, btn_gcs[0],
		     x + width - (indicator_size.width + indicator_spacing.left + indicator_spacing.right) - QT_STYLE->xthickness + 1,
		     y + QT_STYLE->ythickness,
		     x + width - (indicator_size.width + indicator_spacing.left + indicator_spacing.right) - QT_STYLE->xthickness + 1,
		     y + height - QT_STYLE->ythickness - 1);
	
		     
      if (area)
	{
	  gdk_gc_set_clip_rectangle (btn_gcs[0], NULL);
	  gdk_gc_set_clip_rectangle (btn_gcs[4], NULL);
	}
    }
  else
    {
      parent_class->draw_box (style, window, state_type, shadow_type, area,
			      widget, detail, x, y, width, height);
    }

#ifndef GTK2_VER   /* CPD GTK1 ONLY??? */
  if (DETAIL ("slider"))
    {
      const char *slider_detail = detail;
      GtkOrientation orientation = GTK_ORIENTATION_VERTICAL;
      
      if (widget)
	{
	  if (GTK_IS_HSCALE (widget))
	    {
	      slider_detail = "hscale";
	      orientation = GTK_ORIENTATION_HORIZONTAL;
	    }
	  else if (GTK_IS_VSCALE (widget))
	    {
	      slider_detail = "vscale";
	      orientation = GTK_ORIENTATION_VERTICAL;
	    }
	  else if (GTK_IS_HSCROLLBAR (widget))
	    orientation = GTK_ORIENTATION_HORIZONTAL;
	}
      
      real_draw_slider (style, window, state_type, shadow_type,
			area, widget, slider_detail, x, y, width, height, orientation);
    }
#endif
  if (detail && (!strcmp (detail, "paned") || !strcmp (detail+1, "paned")))
    {
      GtkOrientation orientation = GTK_ORIENTATION_HORIZONTAL;

      if (*detail == 'h')
	orientation = GTK_ORIENTATION_VERTICAL;
      
      draw_handle (style, window, state_type, shadow_type,
		   area, widget, detail,
		   x, y, width, height, orientation);
    }
  if(custom_c)
    RELEASE_GCS(btn_gcs);
}

static void
#ifdef GTK2_VER
draw_slider
#else
real_draw_slider
#endif
                 (GtkStyle      *style,
		  GdkWindow     *window,
		  GtkStateType   state_type,
		  GtkShadowType  shadow_type,
		  GdkRectangle  *area,
		  GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
		  gchar         *detail,
		  gint           x,
		  gint           y,
		  gint           width,
		  gint           height,
		  GtkOrientation orientation)
{
  QT_CURVE_STYLE_VAR
  gboolean custom_c = FALSE,
           horiz_scale = DETAIL("hscale");
  GdkGC    *btn_gcs[7],
           *bgnd_gc=NULL;

  if(use_button_color(detail))
  {
    if(QT_CUSTOM_COLOR_BUTTON(detail, style))
    {
      GdkColor btn_colors[7];
      custom_c=TRUE;
      shade_colors(&(style->bg[state_type]), btn_colors);
      GEN_GCS(style, btn_colors, btn_gcs);
    }
    else
      memcpy(btn_gcs, qtcurve_style->button_gc, 7 * sizeof(GdkGC *));
  }
#ifdef GTK2_VER
  g_return_if_fail (GTK_IS_STYLE (style));
#endif  
  g_return_if_fail (window != NULL);
  
  sanitize_size (window, &width, &height);

#ifdef GTK2_VER
  gtk_paint_box (style, window, state_type, shadow_type,
                 area, widget, detail, x, y, width, height);
#endif

 /* Orientation is always vertical with Mozilla, why? Anyway this hack should be OK - as we only draw dashes when slider is
     larger than 23 pixels... */
  orientation=width<height ? GTK_ORIENTATION_VERTICAL : GTK_ORIENTATION_HORIZONTAL;
  if (detail && (0==strcmp ("slider", detail) || horiz_scale || 0==strcmp ("vscale", detail)) &&
      (((horiz_scale || GTK_ORIENTATION_HORIZONTAL==orientation) && width>=20) || height>=20))
      draw_lines(window, x, y, width, height, !(horiz_scale || GTK_ORIENTATION_HORIZONTAL==orientation), 4, 3, btn_gcs, area);

  if(custom_c)
    RELEASE_GCS(btn_gcs);
}

#ifndef GTK2_VER
static void
draw_slider (GtkStyle      *style,
	     GdkWindow     *window,
	     GtkStateType   state_type,
	     GtkShadowType  shadow_type,
	     GdkRectangle  *area,
	     GtkWidget     *widget,
	     gchar         *detail,
	     gint           x,
	     gint           y,
	     gint           width,
	     gint           height,
	     GtkOrientation orientation)
{
  QtCurveStyle *qtcurve_style = style->engine_data;
  
  g_return_if_fail (window != NULL);
  
  sanitize_size (window, &width, &height);

  gtk_paint_box (style, window, state_type, shadow_type,
		 area, widget, detail, x, y, width, height);
  
  real_draw_slider (style, window, state_type, shadow_type,
		    area, widget, detail, x, y, width, height, orientation);
}
#endif

static void 
draw_shadow_gap (GtkStyle       *style,
		 GdkWindow      *window,
		 GtkStateType    state_type,
		 GtkShadowType   shadow_type,
		 GdkRectangle   *area,
		 GtkWidget      *widget,
#ifdef GTK2_VER
                      const
#endif
		 gchar          *detail,
		 gint            x,
		 gint            y,
		 gint            width,
		 gint            height,
		 GtkPositionType gap_side,
		 gint            gap_x,
		 gint            gap_width)
{
  QT_CURVE_STYLE_VAR
  GdkGC *outer_gc = NULL;
  GdkGC *gc1 = NULL;
  GdkGC *gc2 = NULL;
  GdkGC *gc3 = NULL;
  GdkGC *gc4 = NULL;
 
#ifdef GTK2_VER
  g_return_if_fail (GTK_IS_STYLE (style));
#endif 
  g_return_if_fail (window != NULL);
  
  sanitize_size (window, &width, &height);
  
  outer_gc = qtcurve_style->gray_gc[5];
  
  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      return;
    case GTK_SHADOW_IN:
      gc1 = qtcurve_style->gray_gc[5];
      gc2 = qtcurve_style->gray_gc[0];
      gc3 = qtcurve_style->gray_gc[0];
      gc4 = qtcurve_style->gray_gc[5];
      break;
    case GTK_SHADOW_ETCHED_IN:
      gc1 = qtcurve_style->gray_gc[5];
      gc2 = qtcurve_style->gray_gc[0];
      gc3 = qtcurve_style->gray_gc[5];
      gc4 = qtcurve_style->gray_gc[0];
      break;
    case GTK_SHADOW_OUT:
      gc1 = qtcurve_style->gray_gc[5];
      gc2 = qtcurve_style->gray_gc[0];
      gc3 = qtcurve_style->gray_gc[0];
      gc4 = qtcurve_style->gray_gc[5];
      break;
    case GTK_SHADOW_ETCHED_OUT:
      gc1 = qtcurve_style->gray_gc[0];
      gc2 = qtcurve_style->gray_gc[5];
      gc3 = qtcurve_style->gray_gc[0];
      gc4 = qtcurve_style->gray_gc[5];
      break;
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle (outer_gc, area);
      gdk_gc_set_clip_rectangle (gc1, area);
      gdk_gc_set_clip_rectangle (gc2, area);
      gdk_gc_set_clip_rectangle (gc3, area);
      gdk_gc_set_clip_rectangle (gc4, area);
    }
  
  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
    case GTK_SHADOW_IN:
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_IN:
    case GTK_SHADOW_ETCHED_OUT:
      switch (gap_side)
        {
        case GTK_POS_TOP:
          gdk_draw_line (window, gc1,
                         x, y, x, y + height - 1);
          gdk_draw_line (window, gc2,
                         x + 1, y, x + 1, y + height - 2);
          
          gdk_draw_line (window, gc3,
                         x + 1, y + height - 2, x + width - 2, y + height - 2);
          gdk_draw_line (window, gc3,
                         x + width - 2, y, x + width - 2, y + height - 2);
          gdk_draw_line (window, gc4,
                         x, y + height - 1, x + width - 1, y + height - 1);
          gdk_draw_line (window, gc4,
                         x + width - 1, y, x + width - 1, y + height - 1);
          if (gap_x > 0)
            {
              gdk_draw_line (window, gc1,
                             x, y, x + gap_x - 1, y);
              gdk_draw_line (window, gc2,
                             x + 1, y + 1, x + gap_x - 1, y + 1);
            }
          if ((width - (gap_x + gap_width)) > 0)
            {
              gdk_draw_line (window, gc1,
                             x + gap_x + gap_width, y, x + width - 2, y);
              gdk_draw_line (window, gc2,
                             x + gap_x + gap_width, y + 1, x + width - 3, y + 1);
            }
          break;
        case GTK_POS_BOTTOM:
          gdk_draw_line (window, gc1,
                         x, y, x + width - 1, y);
          gdk_draw_line (window, gc1,
                         x, y, x, y + height - 1);
          gdk_draw_line (window, gc2,
                         x + 1, y + 1, x + width - 2, y + 1);
          gdk_draw_line (window, gc2,
                         x + 1, y + 1, x + 1, y + height - 1);
          
          gdk_draw_line (window, gc3,
                         x + width - 2, y + 1, x + width - 2, y + height - 1);
          gdk_draw_line (window, gc4,
                         x + width - 1, y, x + width - 1, y + height - 1);
          if (gap_x > 0)
            {
              gdk_draw_line (window, gc4,
                             x, y + height - 1, x + gap_x - 1, y + height - 1);
              gdk_draw_line (window, gc3,
                             x + 1, y + height - 2, x + gap_x - 1, y + height - 2);
            }
          if ((width - (gap_x + gap_width)) > 0)
            {
              gdk_draw_line (window, gc4,
                             x + gap_x + gap_width, y + height - 1, x + width - 2, y + height - 1);
              gdk_draw_line (window, gc3,
                             x + gap_x + gap_width, y + height - 2, x + width - 2, y + height - 2);
            }
          break;
        case GTK_POS_LEFT:
          gdk_draw_line (window, gc1,
                         x, y, x + width - 1, y);
          gdk_draw_line (window, gc2,
                         x, y + 1, x + width - 2, y + 1);
          
          gdk_draw_line (window, gc3,
                         x, y + height - 2, x + width - 2, y + height - 2);
          gdk_draw_line (window, gc3,
                         x + width - 2, y + 1, x + width - 2, y + height - 2);
          gdk_draw_line (window, gc4,
                         x, y + height - 1, x + width - 1, y + height - 1);
          gdk_draw_line (window, gc4,
                         x + width - 1, y, x + width - 1, y + height - 1);
          if (gap_x > 0)
            {
              gdk_draw_line (window, gc1,
                             x, y, x, y + gap_x - 1);
              gdk_draw_line (window, gc2,
                             x + 1, y + 1, x + 1, y + gap_x - 1);
            }
          if ((width - (gap_x + gap_width)) > 0)
            {
              gdk_draw_line (window, gc1,
                             x, y + gap_x + gap_width, x, y + height - 2);
              gdk_draw_line (window, gc2,
                             x + 1, y + gap_x + gap_width, x + 1, y + height - 2);
            }
          break;
        case GTK_POS_RIGHT:
          gdk_draw_line (window, gc1,
                         x, y, x + width - 1, y);
          gdk_draw_line (window, gc1,
                         x, y, x, y + height - 1);
          gdk_draw_line (window, gc2,
                         x + 1, y + 1, x + width - 1, y + 1);
          gdk_draw_line (window, gc2,
                         x + 1, y + 1, x + 1, y + height - 2);
          
          gdk_draw_line (window, gc3,
                         x + 1, y + height - 2, x + width - 1, y + height - 2);
          gdk_draw_line (window, gc4,
                         x, y + height - 1, x + width - 1, y + height - 1);
          if (gap_x > 0)
            {
              gdk_draw_line (window, gc4,
                             x + width - 1, y, x + width - 1, y + gap_x - 1);
              gdk_draw_line (window, gc3,
                             x + width - 2, y + 1, x + width - 2, y + gap_x - 1);
            }
          if ((width - (gap_x + gap_width)) > 0)
            {
              gdk_draw_line (window, gc4,
                             x + width - 1, y + gap_x + gap_width, x + width - 1, y + height - 2);
              gdk_draw_line (window, gc3,
                             x + width - 2, y + gap_x + gap_width, x + width - 2, y + height - 2);
            }
          break;
        }
    }
  
  if (area)
    {
      gdk_gc_set_clip_rectangle (outer_gc, NULL);
      gdk_gc_set_clip_rectangle (gc1, NULL);
      gdk_gc_set_clip_rectangle (gc2, NULL);
      gdk_gc_set_clip_rectangle (gc3, NULL);
      gdk_gc_set_clip_rectangle (gc4, NULL);
    }
}

static void
draw_hline (GtkStyle     *style,
	    GdkWindow    *window,
	    GtkStateType  state_type,
	    GdkRectangle  *area,
	    GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
	    gchar         *detail,
	    gint          x1,
	    gint          x2,
	    gint          y)
{
  QT_CURVE_STYLE_VAR

#ifdef GTK2_VER
  g_return_if_fail (GTK_IS_STYLE (style));
#endif
  g_return_if_fail (window != NULL);
  
  if (area)
      gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[5], area);
  
  if (DETAIL("label"))
    {
      if (state_type == GTK_STATE_INSENSITIVE)
        gdk_draw_line (window, qtcurve_style->gray_gc[0], x1 + 1, y + 1, x2 + 1, y + 1);   
      gdk_draw_line (window, style->text_gc[state_type], x1, y, x2, y);     
    }
  else
    {
      gdk_draw_line (window, qtcurve_style->gray_gc[5],
		     x1, y, x2, y);
      if (DETAIL ("menuitem"))
	gdk_draw_line (window, qtcurve_style->gray_gc[0],
		       x1, y+1, x2, y+1);
    }
  
  if (area)
      gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[5], NULL);
}

static void
draw_vline (GtkStyle     *style,
	    GdkWindow    *window,
	    GtkStateType  state_type,
	    GdkRectangle  *area,
	    GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
	    gchar         *detail,
	    gint          y1,
	    gint          y2,
	    gint          x)
{
  QT_CURVE_STYLE_VAR

#ifdef GTK2_VER
  g_return_if_fail (GTK_IS_STYLE (style));
#endif  
  g_return_if_fail (window != NULL);
  
  if (area)
      gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[5], area);
  
  gdk_draw_line (window, qtcurve_style->gray_gc[5], x, y1, x, y2 - 1);
 
  if(DETAIL("toolbar"))
      gdk_draw_line (window, qtcurve_style->gray_gc[0], x+1, y1, x+1, y2 - 1);

  if (area)
      gdk_gc_set_clip_rectangle (qtcurve_style->gray_gc[5], NULL);
}

static void 
draw_focus (GtkStyle      *style,
	    GdkWindow     *window,
#ifdef GTK2_VER
	    GtkStateType   state_type,
#endif
	    GdkRectangle  *area,
	    GtkWidget     *widget,
#ifdef GTK2_VER
                      const
#endif
	    gchar         *detail,
	    gint           x,
	    gint           y,
	    gint           width,
	    gint           height)
{
  QT_CURVE_STYLE_VAR
  GdkPoint points[5];
  GdkGC    *gc= qtcurve_style->gray_gc[5];
  gboolean free_dash_list = FALSE;
  gint line_width = 1;
  char *dash_list = (char *)"\1\1";
  gint dash_len;

  if(GTK_IS_EDITABLE(widget))
      return;

#ifdef GTK2_VER
  if (widget)
    {
      gtk_widget_style_get (widget,
			    "focus-line-width", &line_width,
			    "focus-line-pattern", (gchar *)&dash_list,
			    NULL);

      free_dash_list = TRUE;
  }
#else
  if(GTK_IS_OPTION_MENU(widget) || GTK_IS_BUTTON(widget) || GTK_IS_RADIO_BUTTON(widget) || GTK_IS_CHECK_BUTTON(widget))
  {
      if(GTK_IS_OPTION_MENU(widget))
      {
          x+=4;
          y+=4;
          width-=26;
          height-=7;
      }
      else if (!(detail && 0==strcmp(detail, "togglebutton")) && (GTK_IS_RADIO_BUTTON(widget) || GTK_IS_CHECK_BUTTON(widget)))
      {
          x+=18;
          width-=18;
      }
      else
      {
          x+=4;
          y+=4;
          width-=7;
          height-=7;
      }
  }
#endif
  
  sanitize_size (window, &width, &height);
  
  if (area)
    gdk_gc_set_clip_rectangle (gc, area);

  gdk_gc_set_line_attributes (gc, line_width,
			      dash_list[0] ? GDK_LINE_ON_OFF_DASH : GDK_LINE_SOLID,
			      GDK_CAP_BUTT, GDK_JOIN_MITER);


  if (detail && !strcmp (detail, "add-mode"))
    {
      if (free_dash_list)
	g_free (dash_list);
      
      dash_list = (char *)"\4\4";
      free_dash_list = FALSE;
    }

  points[0].x = x + line_width / 2;
  points[0].y = y + line_width / 2;
  points[1].x = x + width - line_width + line_width / 2;
  points[1].y = y + line_width / 2;
  points[2].x = x + width - line_width + line_width / 2;
  points[2].y = y + height - line_width + line_width / 2;
  points[3].x = x + line_width / 2;
  points[3].y = y + height - line_width + line_width / 2;
  points[4] = points[0];

  if (!dash_list[0])
      gdk_draw_lines (window, gc, points, 5);
  else
    {
      /* We go through all the pain below because the X rasterization
       * rules don't really work right for dashed lines if you
       * want continuity in segments that go between top/right
       * and left/bottom. For instance, a top left corner
       * with a 1-1 dash is drawn as:
       *
       *  X X X 
       *  X
       *
       *  X
       *
       * This is because pixels on the top and left boundaries
       * of polygons are drawn, but not on the bottom and right.
       * So, if you have a line going up that turns the corner
       * and goes right, there is a one pixel shift in the pattern.
       *
       * So, to fix this, we drawn the top and right in one call,
       * then the left and bottom in another call, fixing up
       * the dash offset for the second call ourselves to get
       * continuity at the upper left.
       *
       * It's not perfect since we really should have a join at
       * the upper left and lower right instead of two intersecting
       * lines but that's only really apparent for no-dashes,
       * which (for this reason) are done as one polygon and
       * don't to through this code path.
       */
      
      dash_len = strlen (dash_list);
      
      if (dash_list[0])
	gdk_gc_set_dashes (gc, 0, dash_list, dash_len);
      
      gdk_draw_lines (window, gc, points, 3);
      
      /* We draw this line one farther over than it is "supposed" to
       * because of another rasterization problem ... if two 1 pixel
       * unjoined lines meet at the lower right, there will be a missing
       * pixel.
       */
      points[2].x += 1;
      
      if (dash_list[0])
	{
	  gint dash_pixels = 0;
	  gint i;
	  
	  /* Adjust the dash offset for the bottom and left so we
	   * match up at the upper left.
	   */
	  for (i = 0; i < dash_len; i++)
	    dash_pixels += dash_list[i];
      
	  if (dash_len % 2 == 1)
	    dash_pixels *= 2;
	  
	  gdk_gc_set_dashes (gc, dash_pixels - (width + height - 2 * line_width) % dash_pixels, dash_list, dash_len);
	}
      
      gdk_draw_lines (window, gc, points + 2, 3);
    }

  gdk_gc_set_line_attributes (gc, 0, GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER);

  if (area)
    gdk_gc_set_clip_rectangle (gc, NULL);

  if (free_dash_list)
    g_free (dash_list);
}

#ifdef GTK2_VER
static void
draw_resize_grip (GtkStyle       *style,
		  GdkWindow      *window,
		  GtkStateType    state_type,
		  GdkRectangle   *area,
		  GtkWidget      *widget,
		  const gchar    *detail,
		  GdkWindowEdge   edge,
		  gint            x,
		  gint            y,
		  gint            width,
		  gint            height)
{
  QtCurveStyle *qtcurve_style = QTCURVE_STYLE (style);
  g_return_if_fail (GTK_IS_STYLE (style));
  g_return_if_fail (window != NULL);
  
  if (area)
    {
      gdk_gc_set_clip_rectangle (style->light_gc[state_type], area);
      gdk_gc_set_clip_rectangle (style->dark_gc[state_type], area);
      gdk_gc_set_clip_rectangle (style->bg_gc[state_type], area);
    }

  switch (edge)
    {
    case GDK_WINDOW_EDGE_NORTH_WEST:
      /* make it square */
      if (width < height)
	{
	  height = width;
	}
      else if (height < width)
	{
	  width = height;
	}
      break;
    case GDK_WINDOW_EDGE_NORTH:
      if (width < height)
	{
	  height = width;
	}
      break;
    case GDK_WINDOW_EDGE_NORTH_EAST:
      /* make it square, aligning to top right */
      if (width < height)
	{
	  height = width;
	}
      else if (height < width)
	{
	  x += (width - height);
	  width = height;
	}
      break;
    case GDK_WINDOW_EDGE_WEST:
      if (height < width)
	{
	   width = height;
	}
      break;
    case GDK_WINDOW_EDGE_EAST:
      /* aligning to right */
      if (height < width)
	{
	  x += (width - height);
	  width = height;
	}
      break;
    case GDK_WINDOW_EDGE_SOUTH_WEST:
      /* make it square, aligning to bottom left */
      if (width < height)
	{
	  y += (height - width);
	  height = width;
	}
      else if (height < width)
	{
	  width = height;
	}
      break;
    case GDK_WINDOW_EDGE_SOUTH:
      /* align to bottom */
      if (width < height)
	{
	  y += (height - width);
	  height = width;
	}
      break;
    case GDK_WINDOW_EDGE_SOUTH_EAST:
      /* make it square, aligning to bottom right */
      if (width < height)
	{
	  y += (height - width);
	  height = width;
	}
      else if (height < width)
	{
	  x += (width - height);
	  width = height;
	}
      break;
    default:
      g_assert_not_reached ();
    }

  /* Clear background */
  gtk_style_apply_default_background (style, window, FALSE,
				      state_type, area,
				      x, y, width, height);   

  switch (edge)
    {
    case GDK_WINDOW_EDGE_WEST:
    case GDK_WINDOW_EDGE_EAST:
      {
	gint xi;

	xi = x;

	while (xi < x + width)
	  {
	    gdk_draw_line (window,
			   qtcurve_style->gray_gc[0],
			   xi, y,
			   xi, y + height);

	    xi++;
	    gdk_draw_line (window,
			   qtcurve_style->gray_gc[5],
			   xi, y,
			   xi, y + height);

	    xi += 2;
	  }
      }
      break;
    case GDK_WINDOW_EDGE_NORTH:
    case GDK_WINDOW_EDGE_SOUTH:
      {
	gint yi;

	yi = y;

	while (yi < y + height)
	  {
	    gdk_draw_line (window,
			   qtcurve_style->gray_gc[0],
			   x, yi,
			   x + width, yi);

	    yi++;
	    gdk_draw_line (window,
			   qtcurve_style->gray_gc[5],
			   x, yi,
			   x + width, yi);

	    yi+= 2;
	  }
      }
      break;
    case GDK_WINDOW_EDGE_NORTH_WEST:
      {
	gint xi, yi;

	xi = x + width;
	yi = y + height;

	while (xi > x + 3)
	  {
	    gdk_draw_line (window,
			   qtcurve_style->gray_gc[5],
			   xi, y,
			   x, yi);

	    --xi;
	    --yi;

	    gdk_draw_line (window,
			   qtcurve_style->gray_gc[0],
			   xi, y,
			   x, yi);

	    xi -= 3;
	    yi -= 3;
	    
	  }
      }
      break;
    case GDK_WINDOW_EDGE_NORTH_EAST:
      {
        gint xi, yi;

        xi = x;
        yi = y + height;

        while (xi < (x + width - 3))
          {
            gdk_draw_line (window,
                           qtcurve_style->gray_gc[0],
                           xi, y,
                           x + width, yi);                           

            ++xi;
            --yi;
            
            gdk_draw_line (window,
                           qtcurve_style->gray_gc[5],
                           xi, y,
                           x + width, yi);

            xi += 3;
            yi -= 3;
          }
      }
      break;
    case GDK_WINDOW_EDGE_SOUTH_WEST:
      {
	gint xi, yi;

	xi = x + width;
	yi = y;

	while (xi > x + 3)
	  {
	    gdk_draw_line (window,
			   qtcurve_style->gray_gc[5],
			   x, yi,
			   xi, y + height);

	    --xi;
	    ++yi;

	    gdk_draw_line (window,
			   qtcurve_style->gray_gc[0],
			   x, yi,
			   xi, y + height);

	    xi -= 3;
	    yi += 3;
	    
	  }
      }
      break;
      
    case GDK_WINDOW_EDGE_SOUTH_EAST:
      {
        gint xi, yi;

        xi = x;
        yi = y;

        while (xi < (x + width - 3))
          {
            gdk_draw_line (window,
                           qtcurve_style->gray_gc[0],
                           xi, y + height,
                           x + width, yi);                           

            ++xi;
            ++yi;
            
            gdk_draw_line (window,
                           qtcurve_style->gray_gc[5],
                           xi, y + height,
                           x + width, yi);                           
            
            xi += 3;
            yi += 3;
          }
      }
      break;
    default:
      g_assert_not_reached ();
      break;
    }
  
  if (area)
    {
      gdk_gc_set_clip_rectangle (style->light_gc[state_type], NULL);
      gdk_gc_set_clip_rectangle (style->dark_gc[state_type], NULL);
      gdk_gc_set_clip_rectangle (style->bg_gc[state_type], NULL);
    }
}

#endif

static GdkGC *
realize_color (GtkStyle * style,
	       GdkColor * color)
{
  GdkGCValues gc_values;

  gdk_colormap_alloc_color (style->colormap, color,
			    FALSE, TRUE);

  gc_values.foreground = *color;

  return gtk_gc_get (style->depth, style->colormap,
		     &gc_values, GDK_GC_FOREGROUND);
}


static void
#ifdef GTK2_VER
qtcurve_style_realize (GtkStyle * style)
#else
theme_realize_style(GtkStyle * style)
#endif
{
  QT_CURVE_STYLE_VAR

#ifdef GTK2_VER
  parent_class->realize (style);
#endif

  GEN_GCS(style, qtcurve_style->gray, qtcurve_style->gray_gc);
  GEN_GCS(style, qtcurve_style->button, qtcurve_style->button_gc);

  qtcurve_style->spot1_gc = realize_color (style, &qtcurve_style->spot1);
  qtcurve_style->spot2_gc = realize_color (style, &qtcurve_style->spot2);
  qtcurve_style->spot3_gc = realize_color (style, &qtcurve_style->spot3);
}

static void
#ifdef GTK2_VER
qtcurve_style_unrealize (GtkStyle * style)
#else
theme_unrealize_style(GtkStyle * style)
#endif
{
  QT_CURVE_STYLE_VAR
  
  /* We don't free the colors, because we don't know if
   * gtk_gc_release() actually freed the GC. FIXME - need
   * a way of ref'ing colors explicitely so GtkGC can
   * handle things properly.
   */
  RELEASE_GCS(qtcurve_style->gray_gc);
  RELEASE_GCS(qtcurve_style->button_gc);

  gtk_gc_release (qtcurve_style->spot1_gc);
  gtk_gc_release (qtcurve_style->spot2_gc);
  gtk_gc_release (qtcurve_style->spot3_gc);
}

#ifdef GTK2_VER
static void
qtcurve_style_init_from_rc (GtkStyle * style,
			       GtkRcStyle * rc_style)
{
  QtCurveStyle *qtcurve_style = QTCURVE_STYLE (style);

  parent_class->init_from_rc (style, rc_style);

  shade_colors(&qt_settings.colors[COLOR_BUTTON], qtcurve_style->button);
  shade_colors(&qt_settings.colors[COLOR_WINDOW], qtcurve_style->gray);

  shade (&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->spot1, 1.62);
  shade (&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->spot2, 1.05);
  shade (&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->spot3, 0.72);
}

#endif

void 
#ifdef GTK2_VER
qtcurve_style_class_init (QtCurveStyleClass * klass)
#else
qtcurve_initialize_style (GtkStyleClass *style_class,
			     GtkStyleClass *parent)
#endif
{
#ifdef GTK2_VER
  GtkStyleClass *style_class = GTK_STYLE_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);
#else
  parent_class = parent;	/* Save for later use */

  *style_class = *parent;
#endif

#ifdef GTK2_VER
  style_class->realize = qtcurve_style_realize;
  style_class->unrealize = qtcurve_style_unrealize;
  style_class->init_from_rc = qtcurve_style_init_from_rc;
  style_class->draw_resize_grip = draw_resize_grip;
#endif
  style_class->draw_arrow = qtcurve_draw_arrow;
  style_class->draw_tab = draw_tab;
  style_class->draw_shadow = draw_shadow;
  style_class->draw_box_gap = draw_box_gap;
  style_class->draw_extension = draw_extension;
  style_class->draw_handle = draw_handle;
  style_class->draw_box = draw_box;
  style_class->draw_check = draw_check;
  style_class->draw_slider = draw_slider;
  style_class->draw_option = draw_option;
  style_class->draw_shadow_gap = draw_shadow_gap;
  style_class->draw_hline = draw_hline;
  style_class->draw_vline = draw_vline;
  style_class->draw_focus = draw_focus;
#ifdef GTK2_VER
  style_class->draw_layout = draw_layout;
#else
  style_class->draw_string = draw_string;
#endif
}

#ifdef GTK2_VER
static GtkRcStyleClass *parent_rc_class;
GType qtcurve_type_rc_style = 0;
#endif

#ifndef GTK2_VER
static void
qtcurve_rc_data_ref (QtCurveRcStyle *qtcurve_style)
{
  qtcurve_style->refcount++;
}

static void
qtcurve_rc_data_unref (QtCurveRcStyle *qtcurve_style)
{
  qtcurve_style->refcount--;
  if (qtcurve_style->refcount == 0)
    {
      g_free (qtcurve_style);
    }
}
#endif

enum
{
    TOKEN_BORDER_FRAME = G_TOKEN_LAST + 1,
    TOKEN_BORDER_BUTTON,
#ifndef GTK2_VER
    TOKEN_XTHICK,
    TOKEN_YTHICK,
#endif
    TOKEN_TRUE,
    TOKEN_FALSE
};

static struct
{
    gchar              *name;
    guint               token;
}
theme_symbols[] =
{
  { "border_frame",  TOKEN_BORDER_FRAME },
  { "border_button", TOKEN_BORDER_BUTTON },
#ifndef GTK2_VER
  { "xthickness",    TOKEN_XTHICK  },
  { "ythickness",    TOKEN_YTHICK  },
#endif

  { "TRUE",          TOKEN_TRUE },
  { "FALSE",         TOKEN_FALSE },
};

static guint n_theme_symbols = sizeof(theme_symbols) / sizeof(theme_symbols[0]);

#ifndef GTK2_VER
static guint
qtcurve_theme_parse_int(GScanner * scanner,
                int *i)
{
  guint token;

  token = g_scanner_get_next_token(scanner);
  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_EQUAL_SIGN)
    return G_TOKEN_EQUAL_SIGN;

  token = g_scanner_get_next_token(scanner);

  if (token != G_TOKEN_INT && token != G_TOKEN_STRING)
    return G_TOKEN_STRING;
  *i = G_TOKEN_STRING==token ? atoi(scanner->value.v_string) : scanner->value.v_int;

  return G_TOKEN_NONE;
}
#endif

static guint
qtcurve_theme_parse_bool(GScanner * scanner,
                gboolean *b)
{
  guint token;

  token = g_scanner_get_next_token(scanner);
  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_EQUAL_SIGN)
    return G_TOKEN_EQUAL_SIGN;

  token = g_scanner_get_next_token(scanner);

  if (token == TOKEN_TRUE)
    *b = TRUE;
  else if (token == TOKEN_FALSE)
    *b = FALSE;
  else
    return TOKEN_TRUE;

  return G_TOKEN_NONE;
}

static guint
#ifdef GTK2_VER
qtcurve_rc_style_parse (GtkRcStyle *rc_style,
			   GtkSettings  *settings,
			   GScanner   *scanner)
#else
theme_parse_rc_style (GScanner   *scanner,
		      GtkRcStyle *rc_style)
#endif
		     
{
  static GQuark scope_id = 0;
#ifdef GTK2_VER
  QtCurveRcStyle *qtcurve_style = QTCURVE_RC_STYLE (rc_style);
#else
  QtCurveRcStyle *qtcurve_style;
#endif
  guint old_scope,
        token;
  
  /* Set up a new scope in this scanner. */

  if (!scope_id)
    scope_id = g_quark_from_string("qtcurve_theme_engine");

  /* If we bail out due to errors, we *don't* reset the scope, so the
   * error messaging code can make sense of our tokens.
   */
  old_scope = g_scanner_set_scope(scanner, scope_id);

  /* We're ready to go, now parse the top level */

  if (!g_scanner_lookup_symbol(scanner, theme_symbols[0].name))
    {
      int i;

      g_scanner_freeze_symbol_table(scanner);
      for (i = 0; i < n_theme_symbols; i++)
        g_scanner_scope_add_symbol(scanner, scope_id,
                                   theme_symbols[i].name,
                                   GINT_TO_POINTER(theme_symbols[i].token));
      g_scanner_thaw_symbol_table(scanner);
    }

#ifndef GTK2_VER
  qtcurve_style = g_new(QtCurveRcStyle, 1);
  qtcurve_style->refcount = 1;
  qtcurve_style->xthickness=qtcurve_style->ythickness=1;
#endif

  token = g_scanner_peek_next_token(scanner);
  while (token != G_TOKEN_RIGHT_CURLY)
    {
      switch (token)
	{
        case TOKEN_BORDER_FRAME:
          token = qtcurve_theme_parse_bool(scanner, &(global.border_frame));
          break;
        case TOKEN_BORDER_BUTTON:
          token = qtcurve_theme_parse_bool(scanner, &(global.border_button));
          break;
#ifndef GTK2_VER
        case TOKEN_XTHICK:
          token = qtcurve_theme_parse_int(scanner, &(qtcurve_style->xthickness));
          break;
        case TOKEN_YTHICK:
          token = qtcurve_theme_parse_int(scanner, &(qtcurve_style->ythickness));
          break;
#endif
	default:
	  g_scanner_get_next_token(scanner);
	  token = G_TOKEN_RIGHT_CURLY;
	  break;
	}

      if (token != G_TOKEN_NONE)
	return token;

      token = g_scanner_peek_next_token(scanner);
    }

  g_scanner_get_next_token(scanner);

#ifndef GTK2_VER
  rc_style->engine_data = qtcurve_style;
#endif
  g_scanner_set_scope(scanner, old_scope);

  return G_TOKEN_NONE;
}

static void
#ifdef GTK2_VER
qtcurve_rc_style_merge (GtkRcStyle *dest,
			   GtkRcStyle *src)
{
  QtCurveRcStyle *dest_w, *src_w;
  
  parent_rc_class->merge (dest, src);

  if (!QTCURVE_IS_RC_STYLE (src))
    return;
  
  src_w = QTCURVE_RC_STYLE (src);
  dest_w = QTCURVE_RC_STYLE (dest);

}
#else
theme_merge_rc_style (GtkRcStyle *dest,
		      GtkRcStyle *src)
{
  QtCurveRcStyle *src_data = src->engine_data;
  
  if (!dest->engine_data)
    {
      if (src_data)
	{
	  qtcurve_rc_data_ref (src_data);
	  dest->engine_data = src_data;
	}
    }
}
#endif

#ifndef GTK2_VER
static void
set_props (GtkStyle *style)
{
  /* TODO: Change and/or add to these: */
#define OPTION_INDICATOR_WIDTH 7
#define OPTION_INDICATOR_LEFT_SPACING 7
#define OPTION_INDICATOR_RIGHT_SPACING 5
  gtk_style_set_prop_experimental (style, "GtkButton::default_spacing", 6);
  gtk_style_set_prop_experimental (style, "GtkCheckButton::indicator_size", CHECK_SIZE);
  gtk_style_set_prop_experimental (style, "GtkOptionMenu::indicator_width", OPTION_INDICATOR_WIDTH);
  gtk_style_set_prop_experimental (style, "GtkOptionMenu::indicator_left_spacing", OPTION_INDICATOR_LEFT_SPACING);
  gtk_style_set_prop_experimental (style, "GtkOptionMenu::indicator_right_spacing", OPTION_INDICATOR_RIGHT_SPACING);
  gtk_style_set_prop_experimental (style, "GtkPaned::handle_full_size", 1);
  gtk_style_set_prop_experimental (style, "GtkPaned::handle_width", 6);
  gtk_style_set_prop_experimental (style, "GtkRange::always_draw_trough", 1);
  gtk_style_set_prop_experimental (style, "GtkRange::trough_border", 0);
  gtk_style_set_prop_experimental (style, "GtkRange::slider_width", 15);
  gtk_style_set_prop_experimental (style, "GtkRange::stepper_size", 15);
  gtk_style_set_prop_experimental (style, "GtkRange::stepper_spacing", 0);
  gtk_style_set_prop_experimental (style, "GtkSpinButton::shadow_type", GTK_SHADOW_IN);
}

static void
theme_rc_style_to_style(GtkStyle * style,
			GtkRcStyle * rc_style)
{
  static GtkStyleClass *default_class = NULL;
  GtkStyleClass        *class=NULL;
  QtCurveStyle         *qtcurve_style;
  QtCurveRcStyle       *qtcurve_rc_style=rc_style->engine_data;

  if (!default_class)
    {
      default_class = g_new (GtkStyleClass, 1);
      qtcurve_initialize_style (default_class, style->klass);
      default_class->xthickness = 1;
      default_class->ythickness = 1;
    }

  if (qtcurve_rc_style->xthickness!=default_class->xthickness || qtcurve_rc_style->ythickness!=default_class->ythickness)
    {
      class = g_new (GtkStyleClass, 1);
      qtcurve_initialize_style (class, style->klass);
      class->xthickness = qtcurve_rc_style->xthickness;
      class->ythickness = qtcurve_rc_style->ythickness;
    }
  else
      class=default_class;
 
  style->klass = class;
  qtcurve_style = style->engine_data = g_new0 (QtCurveStyle, 1);

  shade_colors(&qt_settings.colors[COLOR_WINDOW], qtcurve_style->gray);
  shade_colors(&qt_settings.colors[COLOR_BUTTON], qtcurve_style->button);

  shade (&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->spot1, 1.62);
  shade (&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->spot2, 1.05);
  shade (&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->spot3, 0.72);
  
  set_props (style);
  qt_set_colours(style, rc_style, 12);
  qt_set_font(style);
}

static void
theme_duplicate_style(GtkStyle * dest,
		      GtkStyle * src)
{
  dest->engine_data = g_memdup (src->engine_data, sizeof (QtCurveStyle));
  
  set_props (dest);
}
#else
/* Create an empty style suitable to this RC style
 */
static GtkStyle *
qtcurve_rc_style_create_style (GtkRcStyle *rc_style)
{
  GtkStyle *style=g_object_new (QTCURVE_TYPE_STYLE, NULL);

  qt_set_colours(style, rc_style, 12);
  qt_set_font(style);

  return style;
}
#endif

#ifndef GTK2_VER
static void
theme_destroy_rc_style(GtkRcStyle * rc_style)
{
  QtCurveRcStyle *rc_data = rc_style->engine_data;
  qtcurve_rc_data_unref (rc_data);
}

static void
theme_destroy_style(GtkStyle * style)
{
  g_free (style->engine_data);
}
#endif

#ifdef GTK2_VER
GType qtcurve_type_style = 0;

static void
qtcurve_style_init (QtCurveStyle * style)
{
}

void
qtcurve_style_register_type (GTypeModule * module)
{
  static const GTypeInfo object_info =
  {
    sizeof (QtCurveStyleClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) qtcurve_style_class_init,
    NULL,			/* class_finalize */
    NULL,			/* class_data */
    sizeof (QtCurveStyle),
    0,				/* n_preallocs */
    (GInstanceInitFunc) qtcurve_style_init,
    NULL
  };

  qtcurve_type_style = g_type_module_register_type (module,
						  GTK_TYPE_STYLE,
						  "QtCurveStyle",
						  &object_info, 0);
}

static void
qtcurve_rc_style_init (QtCurveRcStyle *qtcurve_rc)
{
}

static void
qtcurve_rc_style_class_init (QtCurveRcStyleClass *klass)
{
  GtkRcStyleClass *rc_style_class = GTK_RC_STYLE_CLASS (klass);

  parent_rc_class = g_type_class_peek_parent (klass);

  rc_style_class->parse = qtcurve_rc_style_parse;
  rc_style_class->create_style = qtcurve_rc_style_create_style;
  rc_style_class->merge = qtcurve_rc_style_merge;
  qt_init();
}

void
qtcurve_rc_style_register_type (GTypeModule *module)
{
  static const GTypeInfo object_info =
  {
    sizeof (QtCurveRcStyleClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) qtcurve_rc_style_class_init,
    NULL,           /* class_finalize */
    NULL,           /* class_data */
    sizeof (QtCurveRcStyle),
    0,              /* n_preallocs */
    (GInstanceInitFunc) qtcurve_rc_style_init,
    NULL
  };
  
  qtcurve_type_rc_style = g_type_module_register_type (module,
						     GTK_TYPE_RC_STYLE,
						     "QtCurveRcStyle",
						     &object_info, 0);
}
#endif

G_MODULE_EXPORT void
#ifdef GTK2_VER
theme_init (GTypeModule *module)
#else
theme_init (GtkThemeEngine * engine)
#endif
{
#ifdef GTK2_VER
  qtcurve_rc_style_register_type (module);
  qtcurve_style_register_type (module);
#else
  engine->parse_rc_style = theme_parse_rc_style;
  engine->merge_rc_style = theme_merge_rc_style;
  engine->rc_style_to_style = theme_rc_style_to_style;
  engine->duplicate_style = theme_duplicate_style;
  engine->realize_style = theme_realize_style;
  engine->unrealize_style = theme_unrealize_style;
  engine->destroy_rc_style = theme_destroy_rc_style;
  engine->destroy_style = theme_destroy_style;
  engine->set_background = NULL;

  gdk_rgb_init ();
  qt_init();
#endif
}

G_MODULE_EXPORT void
theme_exit (void)
{
    qt_exit();
}

#ifdef GTK2_VER
G_MODULE_EXPORT GtkRcStyle *
theme_create_rc_style (void)
{
      return GTK_RC_STYLE (g_object_new (QTCURVE_TYPE_RC_STYLE, NULL));
}
#endif

/* The following function will be called by GTK+ when the module
 * is loaded and checks to see if we are compatible with the
 * version of GTK+ that loads us.
 */
G_MODULE_EXPORT const gchar* g_module_check_init (GModule *module);
const gchar*
g_module_check_init (GModule *module)
{
  return gtk_check_version (GTK_MAJOR_VERSION,
			    GTK_MINOR_VERSION,
			    GTK_MICRO_VERSION - GTK_INTERFACE_AGE);
}

